/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.branch.c,v 1.24 1999/04/20 05:18:37 muth Exp $
  Comment:  This file contains the code for the optimizations:
                o useless conditional branches
                o branch forwarding
                o bbl merging
                o fallthru optimization
                
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
@FUN
Assumning that ibbl has a conditional branch reverse
the true and false branch.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblBranchTwist"


static NUMBER Opcode[14] =
{
    I_beq, I_bne,
    I_blbc,I_blbs,
    I_ble, I_bgt,
    I_bge, I_blt,
    I_fbeq,I_fbne,
    I_fble,I_fbgt,
    I_fbge,I_fblt
};


extern void AbblBranchTwist(INDEX ibbl)
{
    INDEX iins;
    INDEX tedg;
    int i;
    
    CASSERT( ABBL_TYPE(ibbl) == BT_BRANCH,("expected BT_BRANCH for bbl %d\n",ibbl));
    iins = ABBL_INS_LAST(ibbl);

    if( AINS_TYPE(iins) != IT_IBR && AINS_TYPE(iins) != IT_FBR )
    {
        ERROR(("bad last instruction in bbl %d\n",ibbl));
        PrintIns(stdout,iins);
    }

    /* change opcode */
    
    for(i=0;i<14;i++)
    {
        if( Opcode[i] == AINS_CODE(iins) )
            break;
    }
    CASSERT(i<14,("branch opcode not found\n"));
    i ^=  1;
    AINS_CODE(iins) = Opcode[i];

    /* swap edges */
    
    tedg = ABBL_EDG_FALSE(ibbl);
    ABBL_EDG_FALSE(ibbl) = ABBL_EDG_TRUE(ibbl);
    ABBL_EDG_TRUE(ibbl) = tedg;

    AEDG_TYPE(ABBL_EDG_TRUE(ibbl)) = ET_TRUE;
    AEDG_TYPE(ABBL_EDG_FALSE(ibbl)) = ET_FALSE;
}

#if 0
/*=======================================================
@OPT
obsolete (is now part of layouter)
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptBranchTwist"

extern BOOL OptBranchTwist(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Branchtwist");
    BOOL   change;
    INDEX  ibbl;
    INDEX  fbbl;
    INDEX  tbbl;
    NUMBER count = 0;
    NUMBER tcount = 0;
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;

        tbbl = AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
        fbbl = AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
               
        tcount++;

        if( tbbl == ibbl ) continue;

#if 0
        if( CONFIG_USE_PROFILE )
        {
            if( ABBL_COUNT(tbbl) == 0 )
            {
                tflow = 0;
                fflow = ABBL_COUNT(ibbl);
            }
            else if( ABBL_COUNT(fbbl) == 0 )
            {
                fflow = 0;
                tflow = ABBL_COUNT(ibbl);
            }
            else if( AbblNpreds(tbbl) == 1 )
            {
                tflow = ABBL_COUNT(tbbl);
                fflow = ABBL_COUNT(ibbl) - tflow;
            }
            else if( AbblNpreds(fbbl) == 1 )
            {
                fflow = ABBL_COUNT(fbbl);
                tflow = ABBL_COUNT(ibbl) - fflow;
            }
            else
            {
                /* we have to guess */
                fflow = (ABBL_COUNT(fbbl)*ABBL_COUNT(ibbl)) /
                        (ABBL_COUNT(fbbl)+ABBL_COUNT(tbbl));
                tflow = (ABBL_COUNT(tbbl)*ABBL_COUNT(ibbl)) /
                        (ABBL_COUNT(fbbl)+ABBL_COUNT(tbbl));
            }
        }
        else
        {
            tflow = 0;
            fflow = 0;
        }
#endif
        
        if( ibbl == fbbl ||
            (ABBL_IS_HOT(ibbl) && !ABBL_IS_HOT(fbbl)) ||
            (!ABBL_IS_HOT(ibbl) && ABBL_IS_HOT(fbbl)) )
        {
            if( verbose )
            {
                fprintf(fp,"\tCC %ld  FC: %ld  TC: %ld  [%s]\n",
                        ABBL_COUNT(ibbl),
                        ABBL_COUNT(fbbl),
                        ABBL_COUNT(tbbl),
                        AFUN_NAME(ABBL_FUN(ibbl)));
            }
            
            count++;
            change = 1;
            
            AbblBranchTwist(ibbl);
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (count>0);

    InfoStatus(fp,"[OptBranchtwist%c] Twisted Conditionals: %d/%d {%d ms}\n",
            CHANGE(change),count,tcount,time);
    
    return 0;   /* there should be only one run of this optimization */
}
#endif

/*=======================================================
@OPT
If both branches of a conditional point to the same basic
block the conditional is useless.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptBranchConditional"

extern BOOL OptBranchConditional(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Cond");
    BOOL   change;    
    INDEX  ibbl;
    NUMBER count = 0;
    NUMBER counth = 0;
    INDEX iins;
    INDEX bt;
    INDEX ft;
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
        
        bt =  AEDG_TAIL(ABBL_EDG_TRUE(ibbl));
        ft =  AEDG_TAIL(ABBL_EDG_FALSE(ibbl));
        
        if( bt != ft ) continue;

        iins = ABBL_INS_LAST(ibbl);
        
        CASSERT( AINS_TYPE(iins) == IT_IBR || AINS_TYPE(iins) == IT_FBR,
                 ("weird\n"));
        
        if( verbose )
        {
            fprintf(fp,"\tUnconditionalizing: ");
            PrintIns(fp,iins);
        }
        
        AinsKill(iins);
        ABBL_TYPE(ibbl) = BT_NORM;
        
        AedgKillFancy(ABBL_EDG_TRUE(ibbl));
        AedgKillFancy(ABBL_EDG_FALSE(ibbl));
        AedgLinkFancy( AedgNewFancy(ibbl,bt,ET_NORM) );
        
        count++;
        if( ABBL_IS_HOT(ibbl) )
            counth++;
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = count >0;

    InfoStatus(fp,"\t[OptCond%c] Unconditionalilized Branches: %d#%d {%d ms}\n",CHANGE(change),count,counth,time);

    return change;
}
/*=======================================================
@FUN
Follow chains of empty basic blocks until a nonempty basic block is found
which is returned.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "FindNewBrTarget"

static INDEX FindNewBrTarget(INDEX ibbl)
{
    NUMBER i;

    if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) return ibbl;
    if( ABBL_TYPE(ibbl) == BT_EXIT ) return ibbl;
    
    for( i=1024; i>=0; i--)
    {
        if( ABBL_HAS_INSS(ibbl) ) return ibbl;
        
        CASSERT( ABBL_TYPE(ibbl) == BT_NORM,("empty bbl is non bt_norm bbl\n"));
        ibbl =  AEDG_TAIL(ABBL_EDG_NORM(ibbl));
    }
    
    WARNING(("potential infinite loop\n"));
    return ibbl;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AedgForward"

static NUMBER AedgForward(INDEX iedg,FILE *fp, BOOL verbose)
{
    INDEX oldtarget = AEDG_TAIL(iedg);
    INDEX newtarget = FindNewBrTarget(oldtarget);
    INDEX xedg;
    
    if( oldtarget != newtarget )
    {
       xedg = AedgNewFancy(AEDG_HEAD(iedg),newtarget,AEDG_TYPE(iedg));
       AedgKillFancy(iedg);
       AedgLinkFancy(xedg);
       if( verbose )
       {
           fprintf(fp,"\t[*] forwarding %d to %d\n",oldtarget,newtarget);
       }
       
       return 1;
    }
    
    return 0;
}

/*=======================================================
@OPT
Forward branches over empty basic blocks.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "OptBranchForwarding"

extern BOOL OptBranchForwarding(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Forward");
    BOOL   change;
    INDEX  ibbl;
    NUMBER count = 0;
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) == BT_NORM )
        {
            count += AedgForward(ABBL_EDG_NORM(ibbl),fp,verbose);
        }
        
        if( ABBL_TYPE(ibbl) == BT_UNCOND )
        {
            count += AedgForward(ABBL_EDG_UNCOND(ibbl),fp,verbose);
        }

        if( ABBL_TYPE(ibbl) == BT_BRANCH )
        {
            count += AedgForward(ABBL_EDG_TRUE(ibbl),fp,verbose);
            count += AedgForward(ABBL_EDG_FALSE(ibbl),fp,verbose);
        }

        if( ABBL_TYPE(ibbl) == BT_CALL )
        {
            if( AedgForward(ABBL_EDG_LINK(ibbl),fp,verbose) )
            {
                INDEX oldcall = ABBL_EDG_CALL(ibbl);
                INDEX newcall = AedgNewFancy(AEDG_HEAD(oldcall),AEDG_TAIL(oldcall),AEDG_TYPE(oldcall));
                count++;
                AedgKillFancy(oldcall);
                AedgLinkFancy(newcall);
            }
            
        }
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = 0;   /* NEEDS WORK */

    InfoStatus(fp,"\t[OptForward%c] Forwarded Branches: %d {%d ms}\n",CHANGE(change),count,time);
    
    return change;
}

/*=======================================================
@OPT
Merge basic blocks which are connected by a normal edge if
the target basic block has no other incoming edges.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptBranchMerge"

extern BOOL OptBranchMerge(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Mergebbl");
    BOOL   change;
    INDEX  iedg;
    INDEX  iins;
    NUMBER count = 0;
    NUMBER counth = 0;
    clock_t time = clock();

    FOREACH_LIVE_EDG(iedg)
    {
        INDEX hd = AEDG_HEAD(iedg);
        INDEX tl = AEDG_TAIL(iedg);

        if( AEDG_TYPE(iedg) != ET_NORM ) continue;
        if( hd == tl ) continue;
        if( ABBL_TYPE(hd) == BT_EXIT ) continue;
        if( ABBL_TYPE(tl) == BT_EXIT ) continue;
        if( ABBL_IS_RELOC(tl) ) continue;
        
        if( AbblNsuccs(hd) != 1 ) continue;
        if( AbblNpreds(tl) != 1 ) continue;

        CASSERT( ABBL_TYPE(hd) == BT_NORM,("weird\n"));
        
        if( verbose )
        {
            fprintf(fp,"\t[*] merging edg %d: %d %d\n",iedg,hd,tl);
#if 0
            PrintBblSuper(stdout,hd);
            PrintBblSuper(stdout,tl);
#endif
            
        }

#if 1        
        if( AEDG_IS_INTER(iedg) )
        {
            INDEX hd_fun = ABBL_FUN(hd);
            INDEX tl_fun = ABBL_FUN(tl);
            
            WARNING(("Merging bbls across function boundaries %d [%s] -> %d [%s]\n",
                     hd_fun,AFUN_NAME(hd_fun),tl_fun,AFUN_NAME(tl_fun) ));
#if 0
            PrintEdgSuper(stdout,iedg);
#endif
        }
#endif
        
        count++;
        if( ABBL_IS_HOT(hd) )
            counth++;
        
        ABBL_TYPE(hd) = ABBL_TYPE(tl);
        
        FOREACH_BBL_INS(tl,iins)
        {
            INDEX jins = AinsDup(iins);
            AINS_BBL(jins) = hd;
            AinsAppend(jins);
        }

        AedgKillFancy(iedg);
        
        AbblMoveSuccEdges(tl,hd);
        
        AbblKill(tl);
#if 0
        if( verbose )
        {
            fprintf(fp,"\t[*] after\n");
            PrintBblSuper(stdout,hd);
            PrintBblSuper(stdout,tl);
            
        }
#endif
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (count>0);

    InfoStatus(fp,"\t[OptMergebbl%c] Merged Bbls pairs: %d#%d {%d ms}\n", CHANGE(change),count,counth,time);
    
    return change;
}

/*=======================================================
@OPT
Propagation of conditionals, ie.:

If there is a conditional branch "if x goto t else goto f"
where t or f are also conditional branches whose outcome is
determined by the first conditional the obvious optimization will performed.
We only catch the simplest case here and find quite a few oportunities.
We can do a lot better, though. See pldi 1997
"Interprocedural Branch Elimination" 
Bofik, Gupta, Soffa
@END
=======================================================*/

static NUMBER table[] =
{
    I_bne,  I_bne,  1,
    I_bne,  I_beq,  0,
    
    I_beq,  I_beq,  1,
    I_beq,  I_bne,  0,
    I_beq,  I_blbc, 1,
    I_beq,  I_blbs, 0,
    I_beq,  I_bgt,  0,
    I_beq,  I_blt,  0,

    I_blbc, I_blbc, 1,
    I_blbc, I_blbs, 0,

    I_blbs, I_blbs, 1,
    I_blbs, I_blbc, 0,
    I_blbs, I_bne,  1,
    I_blbs, I_beq,  0,
    
    I_ble,  I_ble,  1,
    I_ble,  I_bgt,  0,

    I_bgt,  I_bgt,  1,
    I_bgt,  I_ble,  0,
    I_bgt,  I_bne,  1,
    I_bgt,  I_beq,  0,

    I_bge,  I_bge,  1,  
    I_bge,  I_blt,  0,

    I_blt,  I_blt,  1,  
    I_blt,  I_bge,  0,
    I_blt,  I_bne,  1,
    I_blt,  I_beq,  0,
    
    I_fbeq, I_fbeq, 1,      /* we need to be careful with floating point there are positive and negative zeros! */
    I_fbeq, I_fbne, 0,

    I_fbne, I_fbne, 1,
    I_fbne, I_fbeq, 0,
    
    I_fble, I_fble, 1, 
    I_fble, I_fbgt, 0,

    I_fbgt, I_fbgt, 1, 
    I_fbgt, I_fble, 0,

    I_fbge, I_fbge, 1,
    I_fbge, I_fblt, 0,

    I_fblt, I_fblt, 1,
    I_fblt, I_fbgt, 0,

    INVALID,INVALID,INVALID
};







#undef __PROC__
#define __PROC__ "OptCondPropagate"


extern BOOL OptCondPropagate(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Condpropagate");
    BOOL   change;
    INDEX  iedg;
    INDEX  jedg;
    INDEX  ibbl;
    INDEX  pred;
    INDEX  iins_pred;
    INDEX  iins_this;
    NUMBER opcode;
    NUMBER count = 0;
    NUMBER counth = 0;
    NUMBER i;
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_BRANCH ) continue;
        if( AbblNInss(ibbl) != 1 ) continue;
        
        iins_this = ABBL_INS_LAST(ibbl);
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            pred = AEDG_HEAD(iedg);
            
            if( ABBL_TYPE(pred) != BT_BRANCH ) continue;
            iins_pred = ABBL_INS_LAST(pred);
            if( AINS_REG_A(iins_pred) != AINS_REG_A(iins_this) ) continue;
            
            opcode = AINS_CODE(iins_pred);
            
            if( AEDG_TYPE(iedg) == ET_FALSE )
            {
                CASSERT( AEDG_TAIL(ABBL_EDG_FALSE(pred)) == ibbl,("weird\n"));
                
                for(i=0;i<14;i++)
                {
                    if( Opcode[i] == opcode )
                        break;
                }
                CASSERT(i<14,("branch opcode not found\n"));
                i ^=  1;
                opcode = Opcode[i];
            }

            
            for(i=0;VALID(table[i]);i+=3)
            {
                if( table[i] != opcode ) continue;
                if( table[i+1] != AINS_CODE(iins_this)) continue;

                count++;
                if( ABBL_IS_HOT(pred) )
                    counth++;

                if( verbose )
                {
                    fprintf(fp,"\t[*] Before\n");
                    PrintBblSuper(stdout,pred);
                    PrintBblSuper(stdout,ibbl);
                }
                
                jedg = AedgNewFancy(AEDG_HEAD(iedg),
                                    AEDG_TAIL( table[i+2] ? ABBL_EDG_TRUE(ibbl) : ABBL_EDG_FALSE(ibbl)) ,
                                    AEDG_TYPE(iedg));
                AedgKillFancy(iedg);
                AedgLinkFancy(jedg);

                if( verbose )
                {
                    fprintf(fp,"\t[*] After\n");
                    PrintBblSuper(stdout,pred);
                    PrintBblSuper(stdout,ibbl);
                }
            }
        }
    }
    
            
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (count>0);

    InfoStatus(fp,"\t[OptCondpropagate%c] Merged Bbls pairs: %d#%d {%d ms}\n", CHANGE(change),count,counth,time);
        
    return change;
}

/*=======================================================
=======================================================*/








