/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.compress.c,v 1.5 1999/04/18 20:21:55 muth Exp $
  Comment:  This file contains the code for the bbl compression
                
=======================================================*/

#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"


/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsInvalidateEval"

extern void AinsInvalidateEvalInfo(INDEX iins)
{
    AINS_FLAGS(iins) &= ~(FI_EVAL_B|FI_EVAL_B|FI_EVAL_C);
    AINS_A_EVAL(iins) = INVALID;
    AINS_B_EVAL(iins) = INVALID;
    AINS_C_EVAL(iins) = INVALID;
}

/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblIsEquivalent"

static NUMBER AbblSuffixIns(INDEX ibbl, INDEX jbbl)
{
    NUMBER n;
    INDEX iins;
    INDEX jins;
    INDEX iedg;
    INDEX jedg;
    
    if( ABBL_TYPE(ibbl) != ABBL_TYPE(jbbl) ) return 0;

    if( ABBL_TYPE(ibbl) == BT_SWITCH )
    {
        return 0;
    }
    else
    {
        iedg = ABBL_EDG_TRUE(ibbl) ;
        jedg = ABBL_EDG_TRUE(jbbl) ;
        if( VALID(iedg) )
        {
            CASSERT( VALID(jedg),("weird\n"));
            if( AEDG_TAIL(iedg) != AEDG_TAIL(jedg) )
                return 0;
        }

        iedg = ABBL_EDG_FALSE(ibbl) ;
        jedg = ABBL_EDG_FALSE(jbbl) ;
        if( VALID(iedg) )
        {
            CASSERT( VALID(jedg),("weird\n"));
            if( AEDG_TAIL(iedg) != AEDG_TAIL(jedg) )
                return 0;
        }
    }
    
    for( iins = ABBL_INS_LAST(ibbl), jins = ABBL_INS_LAST(jbbl), n=0;
         VALID(iins) && VALID(jins);
         iins = AINS_PREV(iins),jins = AINS_PREV(jins),n+=1)
    {
        if( !AinsEqual(iins,jins) ) break;
    }

    return n;
}

/*=======================================================
@OPT
This looks at basic blocks with identical succesors.
If one basic block is a suffix of the other (instructionwise)
we can delete the suffix from the larger block and make it
fall thru into the other.


The optimization benefits from normalization, especially
bne -> beq, etc.
Register reallocation (backwards) and
schedule normalization.

After it is run there might be empty bbls.

Note, that eval information must be invalidated.

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCompress"


extern BOOL OptCompress(FILE *fp)
{
    BOOL change = 0;
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"Compress");
    INDEX xbbl;
    INDEX ibbl;
    INDEX jbbl;
    INDEX iedg;
    INDEX jedg;
    INDEX iins;
    NUMBER n;
    
    NUMBER icount = 0;
    NUMBER icounth = 0;

    clock_t time = clock();
    
    FOREACH_LIVE_BBL(xbbl)
    {
        
        FOREACH_BBL_PRED(xbbl,iedg)
        {
            if( AEDG_IS_INTER(iedg) ) continue;
            ibbl = AEDG_HEAD(iedg);
            if( ibbl == xbbl ) continue;
            
            FOREACH_BBL_PRED(xbbl,jedg)
            {
                if( AEDG_IS_INTER(jedg) ) continue;
                jbbl = AEDG_HEAD(jedg);
                if( jbbl == xbbl ) continue;
                if( jbbl == ibbl ) continue;
                if( jedg == iedg ) break;

#if 0
                if( ABBL_TYPE(ibbl) == BT_RET ) continue;
#endif
                
                n = AbblSuffixIns(ibbl,jbbl);
                if( n  == 0 ) continue;
                
                if( AbblNInss(ibbl) == n )
                {
                    if( verbose )
                    {
                        printf("\t[*] merging stuff\n");
                        PrintBblSuper(stdout,jbbl);
                        PrintBblSuper(stdout,ibbl);
                    }
                    
                    icount += n;
                    if( ABBL_IS_HOT(ibbl) && ABBL_IS_HOT(jbbl) )
                    {
                        icounth += n;
                    }
                        
                    while( n-- )
                    {
                        AinsKill( ABBL_INS_LAST(jbbl) );
                    }

                    FOREACH_BBL_INS(ibbl,iins)
                    {
                        AinsInvalidateEvalInfo(iins);
                    }
                    
                    
                    AbblKillSuccEdges(jbbl);
                    ABBL_TYPE(jbbl) = BT_NORM;
                    AedgLinkFancy( AedgNewFancy(jbbl,ibbl,ET_NORM) );

                    if( verbose )
                    {
                        PrintBblSuper(stdout,jbbl);
                        PrintBblSuper(stdout,ibbl);
                    }
                    
                }
                else if(  AbblNInss(jbbl) == n )
                {
                    if( verbose )
                    {
                        printf("\t[*] merging stuff\n");
                        PrintBblSuper(stdout,ibbl);
                        PrintBblSuper(stdout,jbbl);
                    }
                    
                    icount += n;
                    if( ABBL_IS_HOT(ibbl) && ABBL_IS_HOT(jbbl) )
                    {
                        icounth += n;
                    }
                        
                    while( n-- )
                    {
                        AinsKill( ABBL_INS_LAST(ibbl) );
                    }

                    FOREACH_BBL_INS(jbbl,iins)
                    {
                        AinsInvalidateEvalInfo(iins);
                    }

                    AbblKillSuccEdges(ibbl);
                    ABBL_TYPE(ibbl) = BT_NORM;
                    AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_NORM) );

                    if( verbose )
                    {
                        PrintBblSuper(stdout,ibbl);
                        PrintBblSuper(stdout,jbbl);
                    }
                }
                else if( n > 0 /* && ABBL_TYPE(ibbl) != BT_RET */ )
                {
                    INDEX join_bbl = AbblNew();
                    CASSERT( ABBL_TYPE(ibbl) != BT_SWITCH,("weird\n"));
                    ABBL_COUNT(join_bbl) = ABBL_COUNT(ibbl) + ABBL_COUNT(jbbl);
                    
                    icount += n;
                    if( ABBL_IS_HOT(ibbl) && ABBL_IS_HOT(jbbl) )
                    {
                        icounth += n;
                    }
#if 0
                    printf("\t[*] QQQQQQQQQQQQQQQQQQQQQ %d\n",n);
                    PrintBblSuper(stdout,ibbl);
                    PrintBblSuper(stdout,jbbl);
#endif                    
                    ABBL_TYPE(join_bbl) = ABBL_TYPE(ibbl);
                    ABBL_FUN(join_bbl) = ABBL_FUN(ibbl);
                    AbblAppendFancy(join_bbl);
                    
                        
                    while( n-- )
                    {
                        iins = AinsDup(ABBL_INS_LAST(ibbl));
                        AINS_BBL(iins) = join_bbl;
                        AinsInvalidateEvalInfo(iins);
                        AinsPrepend(iins);
                        AinsKill( ABBL_INS_LAST(ibbl) );
                        AinsKill( ABBL_INS_LAST(jbbl) );
                    }


                    AbblKillSuccEdges(jbbl);
                    AbblMoveSuccEdges(ibbl,join_bbl);
                    ABBL_TYPE(ibbl) = BT_NORM;
                    AedgLinkFancy( AedgNewFancy(ibbl,join_bbl,ET_NORM) );
                    ABBL_TYPE(jbbl) = BT_NORM;
                    AedgLinkFancy( AedgNewFancy(jbbl,join_bbl,ET_NORM) );
#if 0
                    PrintBblSuper(stdout,ibbl);
                    PrintBblSuper(stdout,jbbl);
                    PrintBblSuper(stdout,join_bbl);
#endif
                }
            }
        }
        
    }
        
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    change = (icount>0);

    InfoStatus(fp,"\t[OptCompress%c] inss %d#%d\n",
               CHANGE(change),icount,icounth);
    
    return change;
}

/*=======================================================
=======================================================*/











