/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.cse.c,v 1.4 1999/06/01 22:04:08 muth Exp $
  Comment:  Common Subexpression Elimination
=======================================================*/


#include "global.h" 
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"

#include <string.h>

/*=======================================================
@FUN

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsFindIdenticalPred"

static INDEX AinsFindIdenticalPred(INDEX iins)
{
    INDEX jins;
    
    for( jins=AINS_PREV(iins); VALID(jins);  jins=AINS_PREV(jins) )
    {
        if( AINS_CODE(iins) != AINS_CODE(jins) ) continue;
        if( AINS_REG_A(iins) != AINS_REG_A(jins) ) continue;
        if( AINS_REG_B(iins) != AINS_REG_B(jins) ) continue;
        if( AINS_A_EVAL(iins) != AINS_A_EVAL(jins) ) continue;
        if( AINS_B_EVAL(iins) != AINS_B_EVAL(jins) ) continue;
        if( AINS_REG_B(iins) == REG_LIT )
            if( AINS_DATA(iins) != AINS_DATA(jins) ) continue;
        return jins;
    }

    return INVALID;
}


/*=======================================================
@OPTIMIZATION
Common Subexpression Elimination

Requires Local Used Def Info
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptCommonSubexpressions"

extern BOOL OptCommonSubexpressions(FILE *fp)
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Commonsubexpressions");
    BOOL change = 0;
    BOOL flag;
    INDEX iins;
    INDEX jins;
    INDEX ibbl;
    NUMBER  icounth=0;
    NUMBER  icount=0;
    NUMBER  fcounth=0;
    NUMBER  fcount=0;
    INT64   weight=0;
 
    REGLIST avail;
    clock_t time = clock();

    FOREACH_LIVE_BBL(ibbl)
    {
        flag = 0;
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( AINS_TYPE(iins) == IT_ILD ) continue;
            if( AINS_TYPE(iins) == IT_FLD ) continue;
            if( AINS_TYPE(iins) == IT_ULD ) continue;
            if( AINS_TYPE(iins) == IT_LLD ) continue;
            if( AINS_TYPE(iins) == IT_LDA ) continue;
            if( AINS_TYPE(iins) == IT_ICM ) continue;
            if( AINS_TYPE(iins) == IT_FCM ) continue;
            if( AINS_TYPE(iins) == IT_FCS ) continue;  /* do not replace copies */
            
            if( AinsHasSideEffect(iins) ) continue;
            jins = AinsFindIdenticalPred(iins);
            if( !VALID(jins) ) continue;
            avail = AinsValueAvailibilityInExtBbl(jins, REG2BIT(AINS_REG_C(jins)), iins, ibbl);

            if( verbose )
            {
                printf("\t[*] jins: %d [%s] iins: %d [%s] avail: %016lx\n",jins,AINS_NAME(jins),iins,AINS_NAME(iins),avail);
            }
            
            flag =1;

            avail &= REG_IS_INT(AINS_REG_C(jins)) ? REGS_INT : REGS_FLT;
            if( avail == REGS_NONE )
            {
                if( AINS_TYPE(iins) == IT_MLL ||  AINS_TYPE(iins) == IT_MLQ || AINS_TYPE(iins) == IT_MLH )
                    POTENTIAL(("value not available ins %s  weight %ld\n",AINS_NAME(iins),ABBL_COUNT(ibbl)));
                continue;
            }

            if( REG_IS_INT(AINS_REG_C(jins)) )
            {
                REG reg = RegsGetFirst(avail);
                AinsMakeLda(iins,reg,AINS_REG_C(iins),0);
                icount++;
                if( ABBL_IS_HOT(ibbl) )
                    icounth++;
                weight += ABBL_COUNT(ibbl);
            }
            else
            {
                REG reg = RegsGetFirst(avail);
                AinsMakeCpys(iins,reg,reg,AINS_REG_C(iins));
                fcount++;
                if( ABBL_IS_HOT(ibbl) )
                    fcounth++;
                weight += ABBL_COUNT(ibbl);
            }
        }

        if( flag && verbose)
        {
            PrintBblSuper(stdout,ibbl);            
        }
    }

    change = icount+fcounth>0;

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[OptCommonsubexpressions%c] int: %d#%d  flt: %d#%d   w: %ld  {ms: %ld}\n",
               CHANGE(change),icount,icounth,fcount,fcounth,weight,time);
        
    return change;
}




