/*=======================================================
  Author:   
  Version:  $Id: opt.inline.jumps.c,v 1.8 1999/06/14 22:50:02 muth Exp $
  Comment:  
            
=======================================================*/


#include "global.h"
#include "flags.h"
#include "functions.h"
#include "opcode.h"
#include "regs.h"

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "OptInlineJumps"     

extern BOOL OptInlineJumps(FILE *fp )
{
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Jumps");
    BOOL   change = 0;
    INDEX dbbl, ifun;
    INDEX iins, ibbl;
    clock_t time = clock();
    NUMBER jcount = 0;
    INDEX target;
    INDEX iins_jmp, iins_add, iins_ldl, iins_ind;
    REG indexReg;
    INDEX prev;
    INDEX i, j;
    NUMBER value;
    INDEX testBbl;
    INDEX iedg;
    REGLIST avail;
    INT64 totalcount;
    

#if 1 
    return 0;
#endif

    if( !CONFIG_HAVE_BBL_PROFILE )
    {
        printf("\t[OptInlineJumps] No Inlining because of lack of profiling information\n");
        return 0;
    }
    

    /* only want to create a short new sequence, so regular inlining
     * will take care of the actual inlining 
     */
    /* WARNING: the code below assume the counts are in sorted order.
     */


    for (i=0; i < aIndirFuns; i++) {
	/* skip this for indir calls with 0 as highest count */
	iins = aIndir[i].iins;
	ibbl = AINS_BBL(iins);

	if (verbose) {
	    printf("iins is %d, ibbls is %d\n", iins, ibbl);
	    printf("address is %lx\n", Ins2Addr(iins));
	}
	/* if we've don't have a switch, this won't work */
	if (aIndir[i].values[0].count <= 0) continue;
	if (ABBL_TYPE(ibbl) != BT_SWITCH) continue;

	/* if the block is killed, or not hot, don't inline (duh!) */
	if (ABBL_IS_KILLED(ibbl)) continue;
	if (!ABBL_IS_HOT(ibbl)) continue;
	prev = ABBL_PREV(ibbl);
	/* now find highest probability switch target */
	jcount ++;

	/* find JMP instructions ==================================*/
	
	iins_jmp = FindTableJmp(ibbl,prev);
	if( !VALID(iins_jmp) )
	{
	    ERROR(("invalid iins_jmp\n"));
	    jcount--;
	    continue;
	}


	/* find ADD instructions =======================*/
	
	iins_add = FindTableAdd(ibbl,prev,iins_jmp);
	if( !VALID(iins_add) )
	{
	    ERROR(("invalid iins_add\n"));
	    jcount--;
	    continue;
	}
        
	/* find LDL instructions ==================================*/
	
	iins_ldl = FindTableLdl(ibbl,prev,iins_add);
	if( !VALID(iins_ldl) )
	{
	    ERROR(("invalid iins_ldl\n"));
	    jcount--;
	    continue;
	}
	
	
	/* find IND instructions ==========================*/
	
	iins_ind = FindTableInd(ibbl,prev,iins_ldl);
	if( !VALID(iins_ind) )
	{
	    ERROR(("invalid iins_ind\n"));
	    jcount--;
	    continue;
	}
    

	/* at this point I've found the index instr, load instr, and
	   add instr */

	/* I essentially need to check the index to see if it's going
	 * to load the correct jump table entry */
	indexReg = AINS_REG_A(iins_ind);
	printf("index register is %d\n", indexReg);
	

	/* compute the needed value of the index Reg (based on 
	   expected target of the jump */
	for (j = 0; j < CONFIG_NINDIR; j++) {
	    /* stop when frequency is less than a third */
	    printf("current jump is executed %ld/[%ld] times\n", aIndir[i].values[j].count, aIndir[i].totalcount);
	    printf("index is %d\n", j);

	    /* want to inline if count is > 1/3 of remaining */
	    totalcount = ABBL_COUNT(ibbl);
	    if (!ABBL_IS_HOT(ibbl)) break;
	    if (aIndir[i].values[j].count < totalcount/3) break;

	    /* if we get the -1 value, then there we don't want to count
	       this as inlined, and we know there weren't enough entries
	       */
	    if (aIndir[i].values[j].addr == -1) {
		printf("not enough entries in table\n");
		jcount--;
		break;
	    }

	    value = aIndir[i].values[j].addr;
	    printf("value is %d\n", value);
	    target = ATAB_TARGET(ABBL_JUMP_TAB(ibbl) + value);
	    printf("target bbl is %d\n", target);
	    if (!VALID(target)) {
		printf("invalid target bbl\n");
		jcount--;
		break;
	    }

	    printf("index value should be %d\n", value);
	    printf("looking for blocks dominated by bbl %d\n", target);
	
	    ifun = ABBL_FUN(target);
	    FOREACH_FUN_BBL(ifun, dbbl) {

		if (Dominates(target, dbbl) ) {
		    printf("Block %d dominates %d\n", target, dbbl);
		}
	    }
	    
	    avail = ~AbblRegsLiveIn(ibbl);
	    
	    /* if we don't have a free register */
	   if (avail == REGS_NONE) {
		/* add the checked value back in */
		if (value) {
		    iins = AinsNew();
		    AinsMakeLda(iins, indexReg, indexReg, value);
		    AINS_BBL(iins) = ibbl;
		    AinsInsertBefore(iins, ABBL_INS_FIRST(ibbl));
		}
		avail = indexReg;
	    }
	    else {
		avail = RegsGetFirst(avail);
	    }
	    
	    printf("we're using register %lx\n", avail);
	    testBbl = AbblNew();
	    ABBL_FUN(testBbl) = ABBL_FUN(ibbl);
	    ABBL_TYPE(testBbl) = BT_BRANCH;
	    ABBL_COUNT(testBbl) = ABBL_COUNT(ibbl);
	    /* change ibbl now */
	    ABBL_COUNT(ibbl) -= aIndir[i].values[j].count;
	    AbblAppendFancy(testBbl);
	    printf("saving %ld instructions\n", aIndir[i].values[j].count * 4);

	    if (value) {
		iins = AinsNew();
		AinsMakeLda(iins, indexReg, avail, 0 - value);
		AINS_BBL(iins) = testBbl;
		AinsAppend(iins);
	    }
	    
	    iins = AinsNew();
	    AinsMakeBne(iins, avail, 0);
	    AINS_BBL(iins) = testBbl;
	    AinsAppend(iins);
	    
	    /* at this point, I should delete the edge from all prev, and
	       add in the edge to testBbl */
	    /* test for which edge to replace */
	    FOREACH_BBL_PRED(ibbl, iedg) {
		prev = AEDG_HEAD(iedg);
		printf("looking at previous bbl %d\n", prev);
		if (ABBL_TYPE(prev) != BT_BRANCH) continue;
		printf("breaking previous branch\n");
		if (AEDG_TAIL(ABBL_EDG_TRUE(prev)) == ibbl) {
		    AedgKillFancy(ABBL_EDG_TRUE(prev));
		    AedgLinkFancy( AedgNewFancy(prev, testBbl, ET_TRUE) );
		}
		else {
		    AedgKillFancy(ABBL_EDG_FALSE(prev));
		    AedgLinkFancy( AedgNewFancy(prev, testBbl, ET_FALSE) );
		}
	    }
	    /* now link the two edges for the test bbl */
	    AedgLinkFancy( AedgNewFancy(testBbl, ibbl, ET_TRUE) );
	    AedgLinkFancy( AedgNewFancy(testBbl, target, ET_FALSE) );
	    
	    printf("added bbl %d, with links to %d[t] and %d[f]\n",
		   testBbl, target, ibbl);
	    PrintFunSuper(stdout, ABBL_FUN(ibbl));
	}
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (jcount>0);

    InfoStatus(fp,"\t[OptInlineJumps%c] Inlined Funs: %d {%d ms}\n",
           CHANGE(change),jcount,time);

    return change;
}


