/*=======================================================
  Author:   R. Muth
  Version:  $Id: opt.regrealloc.c,v 1.10 1999/06/15 21:36:52 muth Exp $
  Comment:  This file contains the register reallocation code
=======================================================*/

#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"
/*=======================================================
@FUN
Given an instruction "use" find the instruction before it in
the same basic block which defines register "reg".
If no such instruction can be found return "INVALID".
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "FindInstructionRegDef"

extern INDEX FindInstructionRegDef(INDEX use, REG reg)
{
    INDEX iins;

    for( iins = AINS_PREV(use); VALID(iins); iins =  AINS_PREV(iins) )
    {
        if( AINS_REG_C(iins) == reg )
        {
            if( AINS_TYPE(iins) != IT_FCM && AINS_TYPE(iins) != IT_ICM && AINS_TYPE(iins) != IT_CST )
                return iins;
        }
    }

    return INVALID;
}

/*=======================================================
@FUN
Given an instruction "def" find the instruction after it in
the same basic block which makes the last use of   "reg".
A conditional move instruction which defines "reg" is not
considered a last use!!! neither is a store conditional

If no such instruction can be found return "INVALID".
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "FindInstructionRegLastUse"

extern INDEX FindInstructionRegLastUse(INDEX def, REG reg)
{
    INDEX lastuse = def;
    INDEX iins;

    /*    CASSERT( AINS_REG_C(def) == reg,("weird\n")); */
     
    for( iins = AINS_NEXT(def); VALID(iins); iins =  AINS_NEXT(iins) )
    {
        if( AINS_REG_A(iins) == reg )
            lastuse = iins;

        if( AINS_REG_B(iins) == reg )
            lastuse = iins;
    
        
        if( AINS_REG_C(iins) == reg )
        {
            if( AINS_TYPE(iins) != IT_FCM && AINS_TYPE(iins) != IT_ICM && AINS_TYPE(iins) != IT_CST )
                return lastuse;
            else
                lastuse = iins;
        }
    }

    if( REG2BIT(reg) & ABBL_OUT(AINS_BBL(def)) )
        return INVALID;
    else
        return lastuse;    
}

/*=======================================================
@FUN
Rename a register within a bbl given the defining instruction
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "RenameRegister"

extern void RenameRegister(INDEX iins_def, REG reg_target)     
{
    INDEX iins;
    REG reg_source;
    
    reg_source = AINS_REG_C(iins_def);
    AINS_REG_C(iins_def) = reg_target;

    for( iins = AINS_NEXT(iins_def); VALID(iins); iins =  AINS_NEXT(iins) )
    {
        if( AINS_REG_A(iins) == reg_source ) AINS_REG_A(iins) = reg_target;
        
        if( AINS_REG_B(iins) == reg_source ) AINS_REG_B(iins) = reg_target;

        if( AINS_REG_C(iins) == reg_source )
        {
            if( AINS_TYPE(iins) != IT_FCM && AINS_TYPE(iins) != IT_ICM && AINS_TYPE(iins) != IT_CST )
                return;
            AINS_REG_C(iins) = reg_target;
        }
    }

    CERROR( REG2BIT(reg_source) & ABBL_OUT(AINS_BBL(iins_def)),("cannot rename all uses of register %d\n",reg_source));
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "MyFindScratchRegs"

extern REGLIST MyFindScratchRegs(INDEX def,INDEX use)
{
    INDEX ibbl = AINS_BBL(def);
    INDEX iins;
    REGLIST scratches;

    if( AINS_BBL(use) != AINS_BBL(def) )
        return REGS_NONE;
    
    scratches = REGS_ALL_BUT_ZERO & ~ABBL_OUT(ibbl);

    if( VALID(use) )
    {
        /* filter out those register that are life after the load instruction */
       
        for( iins = ABBL_INS_LAST(ibbl); iins != use; iins = AINS_PREV(iins) )
        {
            scratches |= AinsDef(iins);
            scratches &= ~AinsUse(iins);
        }
    
        scratches |= AinsDef(use);
        scratches &= ~AinsUse(use);
        
        /* a little optimization here */
        
        if( scratches == REGS_NONE ) return REGS_NONE;
        
        /* filter out those register that are used or defined between the store and load */
        
        for( iins = AINS_PREV(use); iins != def; iins = AINS_PREV(iins) )
        {
            scratches &= ~AinsDef(iins);
            scratches &= ~AinsUse(iins);
            
            /* a little optimization here */
            if( scratches == REGS_NONE ) return REGS_NONE;
        }
        
        scratches &= ~AinsDef(def);
    
        return scratches;
    }
    else
    {
        for( iins = ABBL_INS_LAST(ibbl); iins != def; iins = AINS_PREV(iins) )
        {
            scratches &= ~AinsDef(iins);
            scratches &= ~AinsUse(iins);
            
            /* a little optimization here */
            if( scratches == REGS_NONE ) return REGS_NONE;
        }
        scratches &= ~AinsDef(def);
    
        return scratches;
    }
    
    
}

/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "MyFindScratchRegsInExtBbl"

extern REGLIST MyFindScratchRegsInExtBbl(INDEX def,INDEX use)
{
    INDEX ibbl = AINS_BBL(use);
    INDEX iins;
    REGLIST scratches;

    scratches = REGS_ALL_BUT_ZERO & ~ABBL_OUT(ibbl);
    
    /* filter out those register that are life after the load instruction */
       
    for( iins = ABBL_INS_LAST(ibbl); iins != use; iins = AINS_PREV(iins) )
    {
        scratches |= AinsDef(iins);
        scratches &= ~AinsUse(iins);
    }
    
    scratches |= AinsDef(use);
    scratches &= ~AinsUse(use);

    
    /* a little optimization here */
        
    if( scratches == REGS_NONE ) return REGS_NONE;

    iins = AINS_PREV(use);

    scratches &= ~AinsDef(def);
    
    for(;;)
    {
        for( ; VALID(iins); iins = AINS_PREV(iins) )
        {
            if( iins == def )
            {
                return scratches;
            }
            
            scratches &= ~AinsDef(iins);
            scratches &= ~AinsUse(iins);
            
            /* a little optimization here */
            if( scratches == REGS_NONE ) return REGS_NONE;
        }

        if( AbblNpreds(ibbl) != 1 )
        {
            ERROR(("weird\n"));
            return REGS_NONE;
        }

        ibbl = AEDG_HEAD(ABBL_PRED(ibbl));
        scratches &= ~ABBL_OUT(ibbl);
        iins = ABBL_INS_LAST(ibbl);
    }
    
    return scratches;
}


/*=======================================================
@FUN

find pairs of use->def instructions with a write after read
dependency.

Now we can either rename the use or the write.

Renaming the use to reg r

we need to find the defintion preceeding he use and then
have to rename all the uses associated with the use

Renaming the def to reg r

we need to find the last use associated with the def
have to rename all the uses associated with the use

@END
=======================================================*/

#define NO_LAST_USE  (-666)


#undef __PROC__
#define __PROC__ "AbblRegisterReallocReduceConflicts"

static void AbblRegisterReallocReduceConflicts(INDEX ibbl)
{
    REG oldreg;
    REG newreg;
    REGLIST reglist;
    INDEX iins_def;
    INDEX iins;
    INDEX lastuse[64];
    INDEX  rank[64];
    NUMBER best_so_far;
    NUMBER constraint_start;
    NUMBER n;
    
    SCHEDULEES_USED = 0;
    FOREACH_BBL_INS(ibbl,iins)
    {
        ScheduleeNewFancy(iins, INVALID, REGS_NONE);
    }

    /* create constraints */

    CONSTRAINTS_USED = 0;
    ScheduleesCreateMemoryConstraints();  
    ScheduleesCreateResourceConstraints(0,1);
    constraint_start = CONSTRAINTS_USED;

    ScheduleesCreateRegisterConstraints();

    best_so_far = CountAndMarkRelevantConsProdAndProdProdConstraints();

    ScheduleesMarkInstrWithRelavantConsProdAndProdProdEdges();

    
#if 0
    if( ABBL_IS_HOT(ibbl) )
    {
        printf("BBL: %5d iins %d/%d initial constraints %d\n",ibbl, AbblNMarkedInss(ibbl),AbblNInss(ibbl), best_so_far);
        fflush(stdout);
    }
    
#endif
        
    reglist = ABBL_OUT(ibbl);

    for( oldreg=0; oldreg < 64; oldreg++ )
    {
        if( REG2BIT(oldreg) & reglist )
            lastuse[oldreg] = NO_LAST_USE;
        else
            lastuse[oldreg] = INVALID;
    }
    
    FOREACH_BBL_INS_R(ibbl,iins_def)
    {

        if( best_so_far == 0 ) break;

        oldreg = AINS_REG_C(iins_def);

        if( oldreg != REG_ZERO &&
            oldreg != REG_FZERO &&
            (AINS_FLAGS(iins_def) & F_MARKED) &&
            AINS_TYPE(iins_def) != IT_ICM &&
            AINS_TYPE(iins_def) != IT_FCM &&
            lastuse[oldreg] != NO_LAST_USE )
        {
            reglist = MyFindScratchRegs(iins_def, lastuse[oldreg]);
            
            reglist &=  REG_IS_INT(oldreg) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

            if( reglist == REGS_NONE ) goto after_renaming;

            /* we will overwrite the register soon, to let's do the bookeeping here */
            
            lastuse[oldreg] = INVALID;

            /* we to have find out whether there is a register in reglist that induces
               fewer dependencies than reg, the current register */
               
            /* note that reg is guaranteed to not occur in reglist */

            for( newreg=0; newreg < 64; newreg++ )
                    rank[newreg] = 0xffffff;  /* big */
            
            rank[oldreg] = best_so_far;
#if 0
            printf("trial set %016lx\n",reglist);
#endif
            
            for( newreg=0; newreg < 64; newreg++ )
            {
                if( !(REG2BIT(newreg) & reglist) ) continue;

                RenameRegister(iins_def,newreg);

                ScheduleesRemoveRegisterConstraints();
                CONSTRAINTS_USED = constraint_start;
                     
                ScheduleesCreateRegisterConstraints();

                n = CountAndMarkRelevantConsProdAndProdProdConstraints();
                
#if 0
                printf("trying %d  constraints   base %d    total %d  useless %d\n",newreg,constraint_start, CONSTRAINTS_USED,n );
#endif                                
                rank[newreg] = n;
                if( rank[newreg] < best_so_far )
                    best_so_far = rank[newreg];
            }

            /* Now pick the smallest register that achieves the fewest constraints */
            
            for( newreg=0; newreg < 64; newreg++ )
            {
                if( rank[newreg] != best_so_far ) continue;
                RenameRegister(iins_def,newreg);
#if 0
                printf("decided on reg %d\n",newreg);
#endif

                ScheduleesRemoveRegisterConstraints();
                CONSTRAINTS_USED = constraint_start;

                ScheduleesCreateRegisterConstraints();
                n = CountAndMarkRelevantConsProdAndProdProdConstraints();
                ScheduleesMarkInstrWithRelavantConsProdAndProdProdEdges();
                if( best_so_far != n )
                {
                    printf("STRANGE\n");
                }

                break;
            }

            if( newreg == 64 )
            {
                ERROR(("cannot find right register in bbl %d\n",ibbl));
                printf("best so far: %d\n",best_so_far);
                for( newreg=0; newreg < 32; newreg++ )
                {
                    printf("%2d %d\n",newreg,rank[newreg]);
                }
                
            }
            
        }

    after_renaming:

        oldreg = AINS_REG_C(iins_def);
        
        if( AINS_TYPE(iins_def) != IT_FCM && AINS_TYPE(iins_def) != IT_ICM )
        {
            lastuse[oldreg] = INVALID;
        }

        oldreg = AINS_REG_B(iins_def);

        if( lastuse[oldreg] == INVALID )
            lastuse[oldreg] = iins_def;

        oldreg = AINS_REG_A(iins_def);

        if( lastuse[oldreg] == INVALID )
            lastuse[oldreg] = iins_def;
    }

#if 0
    if( ABBL_IS_HOT(ibbl) )
        printf("BBL: %5d  final constraints: %d\n",ibbl,best_so_far );
#endif    
}


/*=======================================================
@FUN

find pairs of use->def instructions with a write after read
dependency.

Now we can either rename the use or the write.

Renaming the use to reg r

we need to find the defintion preceeding he use and then
have to rename all the uses associated with the use

Renaming the def to reg r

we need to find the last use associated with the def
have to rename all the uses associated with the use

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "AbblRegisterReallocReduceRegisters"

static NUMBER AbblRegisterReallocReduceRegisters(INDEX ibbl)
{
    REG oldreg;
    REG newreg;
    REGLIST reglist;
    INDEX iins_def;
    INDEX lastuse[64];
    NUMBER count = 0;
        
    reglist = ABBL_OUT(ibbl);

    for( oldreg=0; oldreg < 64; oldreg++ )
    {
        if( REG2BIT(oldreg) & reglist )
            lastuse[oldreg] = NO_LAST_USE;
        else
            lastuse[oldreg] = INVALID;
    }
    
    FOREACH_BBL_INS_R(ibbl,iins_def)
    {
        oldreg = AINS_REG_C(iins_def);
        
        if( oldreg != REG_ZERO &&
            oldreg != REG_FZERO &&
            oldreg != REG_SP && /* we do not want stacks to be moved to other regs ! - trouble in add_double in gcc */
            AINS_TYPE(iins_def) != IT_ICM &&
            AINS_TYPE(iins_def) != IT_FCM &&
            AINS_TYPE(iins_def) != IT_CST &&
            lastuse[oldreg] != NO_LAST_USE )
        {
#if 0
            printf("looking for %d %d\n",iins_def,lastuse[oldreg]);
#endif       
            
            reglist = MyFindScratchRegs(iins_def, lastuse[oldreg]);

            reglist &=  REG_IS_INT(oldreg) ?  REGS_INT_BUT_ZERO : REGS_FLT_BUT_ZERO;

            newreg = RegsGetFirst(reglist);

            /* only replace register if it is smaller */
               
            if( VALID(newreg) && newreg < oldreg )
            {
#if 0
                printf("rename %d  %d\n",iins_def,newreg);
#endif           
                RenameRegister(iins_def,newreg);
                lastuse[oldreg] = INVALID;
                count++;
            }
        }

        oldreg = AINS_REG_C(iins_def);
        if( AINS_TYPE(iins_def) != IT_FCM && AINS_TYPE(iins_def) != IT_ICM && AINS_TYPE(iins_def) != IT_CST )
        {
            lastuse[oldreg] = INVALID;
        }

        oldreg = AINS_REG_B(iins_def);

        if( lastuse[oldreg] == INVALID )
            lastuse[oldreg] = iins_def;

        oldreg = AINS_REG_A(iins_def);

        if( lastuse[oldreg] == INVALID )
            lastuse[oldreg] = iins_def;
    }

    return count;
}

/*=======================================================
@OPT
Rename the  registers at the basic block level to reduce
the number of read-write conflicts. this should increase
instruction  parallelism and hence help to improve the quality
of the schedule produced by the scheduler.

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptRegisterReallocTwo"

extern BOOL OptRegisterReallocTwo(FILE *fp)
{
    /*
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Regrealloc");
    */
    BOOL   change;
    INDEX  ibbl;
    clock_t time = clock();
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY ) continue;

#if 0
        if( ABBL_IS_HOT(ibbl) )
            PrintBblSuper(stdout,ibbl);
#endif        
        AbblRegisterReallocReduceConflicts(ibbl);
        AbblRegisterReallocReduceConflicts(ibbl);

#if 0
        if( ABBL_IS_HOT(ibbl) )
            PrintBblSuper(stdout,ibbl);
#endif
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = 1;

    InfoStatus(fp,"\t[OptRegrealloctwo%c] {%d ms}\n",
            CHANGE(change),time);
    
    return 0;   /* there should be only one run of this optimization */
}

/*=======================================================
@OPT
Rename the  registers at the basic block level to reduce
the number of register used

@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptRegisterReallocOne"

extern BOOL OptRegisterReallocOne(FILE *fp)
{
    /*
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Regrealloc");
    */
    BOOL   change;
    INDEX  ibbl;
    NUMBER  rcount = 0;
    NUMBER  rcounth = 0;
    NUMBER  bcount = 0;
    NUMBER  bcounth = 0;
    NUMBER n;

    clock_t time = clock();


    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_FLAGS(ibbl) & FB_PROG_ENTRY ) continue;

#if 0
        if( ABBL_IS_HOT(ibbl) || 1 )
        {
            printf("BEFORE\n");
            PrintBblSuper(stdout,ibbl);
        }
            
#endif   
        n = AbblRegisterReallocReduceRegisters(ibbl);
        if( ABBL_IS_HOT(ibbl) )
        {
            bcounth++;
            rcounth+=n;
        }

        bcount++;
        rcount+=n;
#if 0
        if( ABBL_IS_HOT(ibbl) || 1)
        {
            printf("AFTER %d\n",n);
            PrintBblSuper(stdout,ibbl);
        }
#endif   
    }

    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    
    change = (rcount>0);

    InfoStatus(fp,"\t[OptRegreallocone%c] reallocated regs: %d#%d  changed bbls %d#%d {%d ms}\n",
            CHANGE(change),rcount,rcounth,bcount,bcounth,time);
    
    return 0;   /* there should be only one run of this optimization */
}


/*=======================================================
=======================================================*/


