

/*=======================================================
=======================================================*/

#include "global.h"
#include "regs.h"
#include "functions.h"
#include "flags.h"
#include "opcode.h"
#include "fltconst.h"

/*=======================================================
=======================================================*/

#define ABBL_CLONE(i) ABBL_U1(i).doubleword.d1

#define AINS_USE_DEF_POTENTIAL(i) potential[i]

#define FI_CAN_EVAL_A F_LOCAL_1       
#define FI_CAN_EVAL_B F_LOCAL_2
#define FI_CAN_EVAL_C F_LOCAL_3


/*=======================================================
=======================================================*/

static INDEX AbblClone(INDEX ibbl)
{
    if( VALID(ABBL_CLONE(ibbl)) )
        return ABBL_CLONE(ibbl);
    else
        return ibbl;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "AinsSpecialize"

static void AinsSpecialize(INDEX iins, INDEX point, INT64 value)
{
    INDEX ancestor;

    if( !VALID(iins) ) return;
    if( iins >= CONFIG_NINS ) return;

    AinsTrivialEvals(iins);
        
    if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
    {
        ancestor = AINS_A_EVAL(iins);
        if( ancestor == point )
        {
            AINS_FLAGS(iins) |= FI_EVAL_A;
            AINS_A_EVAL(iins) = value;
        }
        else if( ancestor < CONFIG_NINS )
        {
            AinsSpecialize(ancestor,point,value);
            
            if( AINS_FLAGS(ancestor) & FI_EVAL_C )
            {
                AINS_FLAGS(iins) |= FI_EVAL_A;
                AINS_A_EVAL(iins) = AINS_C_EVAL(ancestor);
            }
        }
        
    }

    if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        ancestor = AINS_B_EVAL(iins);
        if( ancestor == point )
        {
            AINS_FLAGS(iins) |= FI_EVAL_B;
            AINS_B_EVAL(iins) = value;
        }
        else if( ancestor < CONFIG_NINS )
        {
            AinsSpecialize(ancestor,point,value);
            
            if( AINS_FLAGS(ancestor) & FI_EVAL_C )
            {
                AINS_FLAGS(iins) |= FI_EVAL_B;
                AINS_B_EVAL(iins) = AINS_C_EVAL(ancestor);
            }
        }
    }

    if( AINS_FLAGS(iins) & FI_EVAL_C ) return;


    if( (AINS_FLAGS(iins) & FI_EVAL_A) && (AINS_FLAGS(iins) & FI_EVAL_B) )
    {
        AinsEval(iins);
    }
}

/*=======================================================
@FUN
Duplicate a marked region of bbls with entry bbl
return clone of entry point.
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunCloneMarkedRegionBbls"     

static INDEX AfunCloneMarkedBbls(INDEX ifun,INDEX entry, INDEX point, INT64 value, double flowfrac)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX iins;
    INDEX iedg;
    
    CASSERT( ifun == ABBL_FUN(entry),("weird\n"));    
    CASSERT( ABBL_FLAGS(entry) & F_MARKED,("weird\n"));

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        ABBL_CLONE(ibbl) = INVALID;
    }

    /* WARNING(("DO BBLS\n")); */
    
    /* do the duplication for ever bbl */
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !(ABBL_FLAGS(ibbl) & F_MARKED) ) continue;

        CASSERT( ABBL_TYPE(ibbl) != BT_EXIT,("weird bt_exit\n"));
        CASSERT( ABBL_TYPE(ibbl) != BT_JUMP,("weird bt_jump\n"));

        jbbl = AbblDup(ibbl);
        ABBL_FUN(jbbl) = ifun;
        AbblAppendFancy(jbbl);
        ABBL_CLONE(ibbl) = jbbl;
        ABBL_CLONE(jbbl) = INVALID;
    }


    /* WARNING(("DO EDGES\n")); */
    
    /* duplicate edges leaving cloned bbls */
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !VALID(ABBL_CLONE(ibbl)) ) continue;

        /* invariant: link edge before call edge */

        if( ABBL_TYPE(ibbl) == BT_CALL || ABBL_TYPE(ibbl) == BT_PAL )
        {
            iedg = ABBL_EDG_LINK(ibbl);            
            AedgLinkFancy( AedgNewFancy(AbblClone(AEDG_HEAD(iedg)), AbblClone(AEDG_TAIL(iedg)),AEDG_TYPE(iedg)) );
            iedg = ABBL_EDG_CALL(ibbl);
            AedgLinkFancy( AedgNewFancy(AbblClone(AEDG_HEAD(iedg)), AEDG_TAIL(iedg),AEDG_TYPE(iedg)) );
        }
        else
        {
            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                if( AEDG_IS_INTER(iedg) )
                {
                    WARNING(("unexpected inter edge\n"));
                    PrintEdgSuper(stdout,iedg);
                    
                }
                AedgLinkFancy( AedgNewFancy(AbblClone(AEDG_HEAD(iedg)), AbblClone(AEDG_TAIL(iedg)),AEDG_TYPE(iedg)) );
            }
        }

    }

    /* WARNING(("DO COUNTS\n")); */
        
    /* fix up bbl counts */

    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !VALID(ABBL_CLONE(ibbl)) ) continue;
       
        ABBL_COUNT(ABBL_CLONE(ibbl)) = (double) ABBL_COUNT(ibbl) * (1.0 - flowfrac);
        ABBL_COUNT(ibbl) = (double) ABBL_COUNT(ibbl) * flowfrac;
    }

    /* duplicate jump tables and patch the bbl - copied from function cloning */

    /* WARNING(("DO JUMPS\n"));*/
        
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        INDEX itab;
        INDEX jtab;

        if( !VALID(ABBL_CLONE(ibbl)) ) continue;
        if( ABBL_TYPE(ibbl) != BT_SWITCH ) continue;
        jbbl = ABBL_CLONE(ibbl);
        
        CASSERT( VALID(ABBL_JUMP_TAB(ibbl)),("jump tab invalid\n") );
        CASSERT( !VALID(ABBL_JUMP_TAB(jbbl)),("jump tab valid\n") );

#if 0
        /* share jumtabs !!!! */
        ABBL_JUMP_TAB(jbbl) = ABBL_JUMP_TAB(ibbl);
#else
        
        FOREACH_BBL_TAB(ibbl,itab)
        {
            jtab = AtabNew();
            ATAB_TARGET(jtab) = AbblClone(ATAB_TARGET(itab));
            if( !VALID(ABBL_JUMP_TAB(jbbl)) )
                ABBL_JUMP_TAB(jbbl) = jtab;            
        }

        /* add terminator */
        jtab = AtabNew();
        ATAB_TARGET(jtab) = INVALID;

        AbblInstallNewJumpTable(jbbl);
#endif
    }

    /* do eval based on known value for point */
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        if( !VALID(ABBL_CLONE(ibbl)) ) continue;
        
        FOREACH_BBL_INS(ibbl,iins)
        {
            AinsSpecialize(iins,point,value);
        }
    }

    return ABBL_CLONE(entry);    
}



/*=======================================================
@FUN

@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AfunMarkUsesTransitive"

extern BOOL AfunMarkUsesTransitive(INDEX ifun, INDEX entry, INDEX point)
{
    INDEX ibbl;
    INDEX iins;
    INDEX def;
    INDEX pred;
    INDEX iedg;
    BOOL flag;

    /* mark dependend instructions */
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            AINS_FLAGS(iins) &= ~(FI_CAN_EVAL_A|FI_CAN_EVAL_B|FI_CAN_EVAL_C);
        }
        
    }

    do
    {
        flag = 0;
        
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            FOREACH_BBL_INS(ibbl,iins)
            {
                if( iins == point ) continue;

                if( !(AINS_FLAGS(iins) & (FI_EVAL_A|FI_CAN_EVAL_A)) )
                {
                    def = AINS_A_EVAL(iins);
                    if( def == point )
                    {
                        AINS_FLAGS(iins) |= FI_CAN_EVAL_A;
                    }
                    if( def >= 0 && def < CONFIG_NINS && (AINS_FLAGS(def) & FI_CAN_EVAL_C)  )
                    {
                        AINS_FLAGS(iins) |= FI_CAN_EVAL_A;
                    }
                }

                if( !(AINS_FLAGS(iins) & (FI_EVAL_B|FI_CAN_EVAL_B)) )
                {
                    def = AINS_B_EVAL(iins);
                    if( def == point )
                    {
                        AINS_FLAGS(iins) |= FI_CAN_EVAL_B;
                        
                    }
                    if( def >= 0 && def < CONFIG_NINS && (AINS_FLAGS(def) & FI_CAN_EVAL_C)  )
                    {
                        AINS_FLAGS(iins) |= FI_CAN_EVAL_B;
                    }
                }
                
                if( AINS_FLAGS(iins) & (FI_EVAL_C|FI_CAN_EVAL_C) ) continue;

                /* we can usually NOT evaluate loads even if we know all arguments */
                   
                if( AINS_TYPE(iins) == IT_ILD ||
                    AINS_TYPE(iins) == IT_FLD ||
                    AINS_TYPE(iins) == IT_ULD ||
                    AINS_TYPE(iins) == IT_LLD ||
                    AINS_TYPE(iins) == IT_IST ||
                    AINS_TYPE(iins) == IT_FST ||
                    AINS_TYPE(iins) == IT_UST ||
                    AINS_TYPE(iins) == IT_CST ) continue;
                    
                if(  !(AINS_FLAGS(iins) & (FI_EVAL_A|FI_CAN_EVAL_A)) ) continue;
                if(  !(AINS_FLAGS(iins) & (FI_EVAL_B|FI_CAN_EVAL_B)) ) continue;
                if(  !(AINS_FLAGS(iins) & (FI_CAN_EVAL_A|FI_CAN_EVAL_B)) ) continue;
                
                AINS_FLAGS(iins) |= FI_CAN_EVAL_C;
                flag = 1;
            }
        }
    }while( flag );


    /* mark bbls */
    
    FOREACH_FUN_BBL(ifun,ibbl)
    {
        flag = 0;
        ABBL_FLAGS(ibbl) &= ~F_MARKED;
        FOREACH_BBL_INS(ibbl,iins)
        {
            /* SAW!!!! for m88ksim set to 0 */
#if 1
            if( !(AINS_FLAGS(iins) & (FI_CAN_EVAL_A|FI_CAN_EVAL_B|FI_CAN_EVAL_C)) ) continue;
#else
            if( !(AINS_FLAGS(iins) & (FI_CAN_EVAL_C)) ) continue;
#endif
            flag = 1;
            printf("\t[*] bbl %6d (%9ld)  ins %7d  (%-10s) %c%c%c\n",
                   ibbl,ABBL_COUNT(ibbl),
                   iins,AINS_NAME(iins),
                   AINS_FLAGS(iins) & FI_CAN_EVAL_A ? 'A' : '_',
                   AINS_FLAGS(iins) & FI_CAN_EVAL_B ? 'B' : '_',
                   AINS_FLAGS(iins) & FI_CAN_EVAL_C ? 'C' : '_');
        }

        if( flag )
        {
            ABBL_FLAGS(ibbl) |= F_MARKED;
        }
        
    }

    ABBL_FLAGS(entry) |= F_MARKED;
    
    /* mark more bbls */
    
    do
    {
        flag = 0;

        FOREACH_FUN_BBL_R(ifun,ibbl)
        {
            if( ibbl == entry ) continue;
            if( ABBL_TYPE(ibbl) == BT_EXIT) continue;
            if( !(ABBL_FLAGS(ibbl) & F_MARKED) ) continue;
            
            FOREACH_BBL_PRED(ibbl,iedg)
            {
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN ) continue;
                if( AEDG_TYPE(iedg) == ET_FUN_CALL ) continue;
                if( AEDG_TYPE(iedg) == ET_HELL ) continue;
                pred = AEDG_HEAD(iedg);
                if( ABBL_FUN(pred) != ifun )
                {
                    ERROR(("unexpected interproc edge\n"));
                    PrintEdgSuper(stdout,iedg);
                }
                if( ABBL_FLAGS(pred) & F_MARKED ) continue;
                ABBL_FLAGS(pred) |= F_MARKED;
                flag = 1;
            }
        }        
    }
    while( flag );

    if( AFUN_BBL_FIRST(ifun) != entry && ABBL_FLAGS(AFUN_BBL_FIRST(ifun)) & F_MARKED )
    {
        ERROR(("fun entry has been marked\n"));
        return 0;
    }

    return 1;
}

/*=======================================================
@FUN
@END
=======================================================*/

extern INT64 TestCost(INT64 value)
{
    INT64 base_cost = 16;
    
    if( value == 0 )
    {
        /* no extra cost */
    }
    if( value == DOUBLE_NZERO )
    {
        base_cost += 4;
    }
    if( value > 0 && value <= 255 )
    {
        base_cost += 2;
    }
    if( value >= -(1<<15) && value < (1<<15) )
    {
        base_cost += 4;
    }
    else
    {
        base_cost += 16;
   }

    return base_cost;
}

/*=======================================================
@FUN
Insert a test for a given value right after the given instruction
Requires Liveness information
@END
=======================================================*/

#undef __PROC__
#define __PROC__ "AinsInsertTestAfter"

static INDEX  AinsInsertTestAfter(INDEX iins, REG reg, INT64 value)
{
    INDEX ifun;
    INDEX jins;
    INDEX ibbl;
    INDEX jbbl;

    
    ibbl = AINS_BBL(iins);
    ifun = ABBL_FUN(ibbl);

    
    /* split  root */
    
    jbbl = AbblNew();
    ABBL_TYPE(jbbl) = ABBL_TYPE(ibbl);
    ABBL_FUN(jbbl) = ifun;
    ABBL_COUNT(jbbl) = ABBL_COUNT(ibbl);
    ABBL_OUT(jbbl) = ABBL_OUT(ibbl);
    ABBL_OUT(ibbl) =  AinsRegsLiveAfter(iins);

    /* we are careful to not duplicate the instructions at the tail !*/

    if( ABBL_INS_LAST(ibbl) != iins )
    {
        ABBL_INS_FIRST(jbbl) = AINS_NEXT(iins);
        ABBL_INS_LAST(jbbl) = ABBL_INS_LAST(ibbl);
        ABBL_INS_LAST(ibbl) = iins;
        jins = AINS_NEXT(iins);
        AINS_NEXT(iins) = INVALID;
        AINS_PREV(jins) = INVALID;
    
        FOREACH_BBL_INS(jbbl,jins)
            AINS_BBL(jins) = jbbl;
    }

    AbblInsertAfter(jbbl,ibbl);

    
    if( REG_IS_INT(reg) )
    {
        if( value == 0 )
        {
            iins = AinsNew();
            AinsMakeBne(iins,reg, 0);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
        }
        else if( REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl) )
        {
            REG reg_scratch = RegsGetFirst(REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl));
            if( value > 0 && value < 256 )
            {
                iins = AinsNew();
                AinsMakeCmpeq(iins,reg, REG_LIT, reg_scratch, value);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);

                iins = AinsNew();
                AinsMakeBeq(iins,reg_scratch, 0);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);

            }
            else if( value >= -(1<<15) && value < (1<<15) )
            {
                iins = AinsNew();
                AinsMakeLda(iins,reg, reg_scratch, -value);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);

                iins = AinsNew();
                AinsMakeBne(iins,reg_scratch, 0);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);

            }
#if 0
           
            else if( LocateSection(value) != SysGetTextSection() && (REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl)) )
            {
                REG int_scratch = RegsGetFirst(REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl));
                INDEX icon = AconNew();
                ADDRESS address = AsecConstSectionAddressData64(icon);
                NUMBER hi;
                NUMBER lo;

                HURRAH(("Specializing Wide Int Constant: %016lx\n",value));
                            
                AsecConstSectionWriteData64(icon,value);

                address -= SysGetGlobalPointer();
                hi = address >> 16;
                lo = address & 0xffff;
                if( lo & 0x8000 )
                {
                    lo |= 0xffff0000;
                    hi += 1;
                }
                CASSERT( ((long)hi << 16) + (long) lo == address,("bad hi/lo address splitt computation\n"));
                
                iins = AinsNew();
                AinsMakeLdah(iins,REG_GP, int_scratch, hi);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);
                
                iins = AinsNew();
                AinsMakeLdq(iins, int_scratch, int_scratch, lo);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);
                
                iins = AinsNew();
                AinsMakeCmpeq(iins, reg, int_scratch, int_scratch,0);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);
                
                iins = AinsNew();
                AinsMakeBeq(iins,int_scratch, 0);
                AINS_BBL(iins) = ibbl;
                AinsAppend(iins);
            }
#endif
            else
            {
                goto failure;
            }
                            
        }
        else
        {
            iins = AinsNew();
            AinsMakeLda(iins,reg, reg, -value);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);

            iins = AinsNew();
            AinsMakeBne(iins,reg, 0);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);

            /* compensation stuff */
            
            jins = AinsNew();
            AinsMakeLda(jins,reg, reg, value);
            AINS_BBL(jins) = jbbl;
            AinsPrepend(jins);
        }
        
    }
    else
    {
        if( value == 0 )
        {
            iins = AinsNew();
            AinsMakeFbne(iins,reg, 0);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
        }
        else if( (value == DOUBLE_PONE  || value == DOUBLE_PHALF ||  value == DOUBLE_PTWO) && 
                 (REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl))  && (REGS_FLT_BUT_ZERO & ~ABBL_OUT(ibbl)) )
        {
            REG int_scratch = RegsGetFirst(REGS_INT_BUT_ZERO & ~ABBL_OUT(ibbl));
            REG flt_scratch = RegsGetFirst(REGS_FLT_BUT_ZERO & ~ABBL_OUT(ibbl));
            INDEX icon = AconNew();
            ADDRESS address = AsecConstSectionAddressData64(icon);
            NUMBER hi;
            NUMBER lo;

            HURRAH(("Specializing Wide Floating Point Constant: %016lx\n",value));

            AsecConstSectionWriteData64(icon,value);

            address -= SysGetGlobalPointer();
            hi = address >> 16;
            lo = address & 0xffff;
            if( lo & 0x8000 )
            {
                lo |= 0xffff0000;
                hi += 1;
            }
            CASSERT( ((long)hi << 16) + (long) lo == address,("bad hi/lo address splitt computation\n"));

            iins = AinsNew();
            AinsMakeLdah(iins,REG_GP, int_scratch, hi);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);

            iins = AinsNew();
            AinsMakeLdt(iins, int_scratch, flt_scratch, lo);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);

            iins = AinsNew();
            AinsMakeCmpteq(iins, reg, flt_scratch, flt_scratch);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);

            iins = AinsNew();
            AinsMakeFbeq(iins,flt_scratch, 0);
            AINS_BBL(iins) = ibbl;
            AinsAppend(iins);
        }
        else
        {
            goto failure;
        }
    }
    
    ABBL_TYPE(ibbl) = BT_BRANCH;

    AbblMoveSuccEdges(ibbl,jbbl);

    AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_TRUE));
    AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_FALSE));
    
    return jbbl;

failure:
    POTENTIAL(("Could not perform specialization\n"));

    
    ABBL_TYPE(ibbl) = BT_NORM;

    AbblMoveSuccEdges(ibbl,jbbl);
    AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_NORM));
#if 0    
    PrintBblSuper(stdout,ibbl);
    PrintBblSuper(stdout,jbbl);
#endif
    return -1;
}



/*=======================================================
@FUN
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptValProfSpecialize"


extern BOOL OptValProfSpecialize(FILE *fp )
{
    INDEX ifun;
    double frac;
    INDEX point;
    INDEX ibbl;
    INDEX jbbl;
    INDEX tbbl;
    INDEX iins;
    INDEX ivpf;
    INDEX jvpf;
    REG reg;
    double flow_frac;
    BOOL change = 0;
    INT64 value;
    NUMBER nins = CONFIG_UINS;
    NUMBER nbbl = CONFIG_UBBL;
    NUMBER nspecialized = 0;
    
    clock_t time = clock();
#if 0
    BOOL verbose = HasOption(PRINT_OPT_VERBOSE,"valprofspecialize");
#endif
     
#if 1
    WARNING(("staring specialization %d profiles loaded\n",CONFIG_NVALUES));
#endif
    
    /* compute benefits */

    FOREACH_VPF(ivpf)
    {
        point = AVPF_MAGIC(ivpf);
        if( point > CONFIG_NINS + 64 * CONFIG_UBBL)
        {
            AVPF_BENEFIT(ivpf) = -1.0;
            continue;
        }

        if( point <= CONFIG_NINS )
        {
            ibbl = AINS_BBL(point);
            iins = point;
            reg = AINS_REG_C(iins);
        }
        else 
        {
            ibbl = (point - CONFIG_NINS) / 64;
            iins = INVALID;
            reg =  (point - CONFIG_NINS) % 64;
        }
        
        if( reg == REG_ZERO || reg == REG_FZERO || reg == REG_SP || reg == REG_GP )
        {
            AVPF_BENEFIT(ivpf) = -1.0;
            continue;
        }
        
        flow_frac =  (double) AVPF_FIRST_COUNT(ivpf) / (2.0+(double) AVPF_COUNT(ivpf));

        if( flow_frac < 0.5 )
        {
            AVPF_BENEFIT(ivpf) = -1.0;
            continue;
        }

#if 0
        if( ibbl > PRINT_OPT_NUM )
        {
            AVPF_BENEFIT(ivpf) = -1.0;
            continue;
        }
        
#endif      

        if( ABBL_IS_KILLED(ibbl) )
        {
            AVPF_BENEFIT(ivpf) = -1.0;
            ERROR(("possible outdated profile detected\n"));
            continue;
        }

        if( VALID(iins) )
        {
            if( AINS_IS_KILLED(iins) || (AINS_FLAGS(iins) & FI_EVAL_C) )
            {
                AVPF_BENEFIT(ivpf) = -1.0;
                ERROR(("possible outdated profile detected iins %d ibbl %d reg %d\n",iins,ibbl,reg));
                continue;
            }
        }

        
        AVPF_BENEFIT(ivpf) = (double) AINS_USE_DEF_POTENTIAL(point) * flow_frac
                                 - (double) TestCost(AVPF_FIRST_VALUE(ivpf)) * (double) ABBL_COUNT(ibbl);

        if( AVPF_BENEFIT(ivpf) < CONFIG_COUNT )
        {
            AVPF_BENEFIT(ivpf) = -1.0;
            continue;
        }
        
#if 1
        printf("magic: %ld iins: %d ibbl: %d  (c:%ld) value: %016lx   pot: %ld   fst: %ld  tot: %ld  %ld\n",
               AVPF_MAGIC(ivpf),iins, ibbl,ABBL_COUNT(ibbl),  AVPF_FIRST_VALUE(ivpf),
               AINS_USE_DEF_POTENTIAL(point),  AVPF_FIRST_COUNT(ivpf),
               AVPF_COUNT(ivpf), (INT64) AVPF_BENEFIT(ivpf));

#endif
    }
    
    /* remove dependent profiling point */

    FOREACH_VPF(ivpf)
    {
        if( AVPF_BENEFIT(ivpf) < 0.0 ) continue;
        
        for(jvpf=0;jvpf<ivpf;jvpf++)
        {
#if 0
            WARNING(("%d (%ld) %d (%ld)\n",jvpf,AVPF_MAGIC(jvpf),ivpf,AVPF_MAGIC(ivpf)));
#endif
            if( AVPF_BENEFIT(jvpf) < 0.0 ) continue;
            if( AinsUseDefIsAncestor(AVPF_MAGIC(ivpf),AVPF_MAGIC(jvpf)) ||
                AinsUseDefIsAncestor(AVPF_MAGIC(jvpf),AVPF_MAGIC(ivpf)) )
            {
#if 0           
                WARNING(("ANCESTORS %d (%ld) %d (%ld)\n",jvpf,AVPF_MAGIC(jvpf),ivpf,AVPF_MAGIC(ivpf)));
#endif
                if( AVPF_BENEFIT(ivpf) > AVPF_BENEFIT(jvpf) )
                    AVPF_BENEFIT(jvpf)  = -1.0;
                else
                    AVPF_BENEFIT(ivpf)  = -1.0;
            }
        }
    }

    
    /* do it the transformation */

    FOREACH_VPF(ivpf)
    {
        if( AVPF_BENEFIT(ivpf) < 0.0 ) continue;
        point = AVPF_MAGIC(ivpf);
        value = AVPF_FIRST_VALUE(ivpf);
        
        if( point <= CONFIG_NINS )
        {
            ibbl = AINS_BBL(point);
            iins = point;
            reg = AINS_REG_C(iins);
        }
        else 
        {
            ibbl = (point - CONFIG_NINS) / 64;
            iins = INVALID;
            reg =  (point - CONFIG_NINS) % 64;
        }
        
        
        ifun = ABBL_FUN(ibbl);
        
        frac = (double) AVPF_FIRST_COUNT(ivpf) / (double) AVPF_COUNT(ivpf);

        WARNING(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
        fprintf(fp,"\t[*] magic %10ld   benefit %10ld  value %016lx  ibbl %6d  reg: %2d (frac: %5.3f)\n",
               AVPF_MAGIC(ivpf),(INT64) AVPF_BENEFIT(ivpf),value,ibbl,reg,frac  );

        if( !VALID(iins) )
        {
            iins = AinsNew();
            AinsMakeLda(iins,REG_ZERO,REG_ZERO,0);
            AINS_BBL(iins) = ibbl;
            AinsPrepend(iins);
        }

        jbbl = AinsInsertTestAfter(iins,reg,value);
        if( !VALID(jbbl) )
        {
            AfunMarkUsesTransitive(ifun, ibbl, point);
            FOREACH_FUN_BBL(ifun,tbbl)
            {
                if( ABBL_FLAGS(tbbl) & F_MARKED ) 
                    PrintBblSuper(stdout,tbbl);
            }
            WARNING(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
            continue;
        }
        
        WARNING(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
        PrintBblSuper(stdout,ibbl);
        PrintBblSuper(stdout,jbbl);
        WARNING(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));

        
        /* WARNING(("MARKING TRANSITIV USES\n")); */
        
        if( !AfunMarkUsesTransitive(ifun, jbbl, point) )
        {
            WARNING(("transitive uses failed\n"));
            continue;
        }
        

        /* WARNING(("MARKING TRANSITIV USES DONE\n")); */
#if 0
        FOREACH_FUN_BBL(ifun,tbbl)
        {
            if( ABBL_FLAGS(tbbl) & F_MARKED ) 
                PrintBblSuper(stdout,tbbl);
        }
        WARNING(("@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n"));
#endif   

        /* WARNING(("CLONING\n"));*/
        
        jbbl = AfunCloneMarkedBbls(ifun,jbbl, point, value, frac);

        if( !VALID(jbbl) )
        {
            WARNING(("Cloning failed\n"));
            continue;
        }

        /* WARNING(("CLONING DONE\n"));*/
                
        nspecialized++;
        
        FOREACH_FUN_BBL(ifun,tbbl)
        {
            if( !VALID(ABBL_CLONE(tbbl)) ) continue;
            PrintBblSuper(stdout,tbbl);
        }
        
        AedgKillFancy(ABBL_EDG_TRUE(ibbl));
        AedgLinkFancy( AedgNewFancy(ibbl,jbbl,ET_TRUE));
    }
    

    
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;
    nins = CONFIG_UINS - nins;
    nbbl = CONFIG_UBBL - nbbl;
    change = (nspecialized>0);

    InfoStatus(fp, "\t[OptValprofspecialize%c] Specialized %d  Extra ins %d Extra bbls %d {%d ms}\n",
	       CHANGE(change),nspecialized,nins,nbbl,time);


    
    return change;
}

/*=======================================================
=======================================================*/

