/*=======================================================
  Author:   R. Muth
  Version:  $Id: optimize.c,v 1.89 1999/07/23 00:34:16 muth Exp $
  Comment:  This file contains the optimization code
=======================================================*/


#include "global.h"
#include "flags.h"
#include "opcode.h"
#include "functions.h"
#include "regs.h"

#define NO_CODE_GROWTH 0


/*=======================================================
@FUNCTION
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeLoadStores"

static BOOL OptimizeLoadStores(FILE *fp)
{
    BOOL flag = 0;
    BOOL cont;

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preuseless") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Useless") )
    {
        do
        {
            cont = 0;
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            cont = OptKillUseless(fp);
            flag |= cont;
        }
        while( cont );
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postuseless") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/
#if 0
    if( HasOption(PRINT_OPT_WHEN,"Prepleafsave") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Leafsave") )
    {
        ComputeSavedChangedRegisters(fp);
        ComputeLiveness(fp);
        flag |= OptLeafSaveRegisters(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"PostLeafsave") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif

    /***************************************************/    
#if 0
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preloop") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Loop") )
    {
        OptRemoveUnreachableCode(fp);
        ComDomsAndLoops(fp);
        ComStackInfo(fp);
        ComUseDefDagLoadStoreExtBbl(fp);
        ComAliasAuxInfo(fp);
            
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComputeFunctionSideEffects(fp);
 
        flag |= OptMoveLoopInvariantCode(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postloop") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif
    
    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preavoidloads") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Avoidloads") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComStackInfo(fp);
        ComUseDefDagLoadStoreExtBbl(fp);
        ComAliasAuxInfo(fp);

        flag |= OptAvoidLoads(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postavoidloads") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

#if 0
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preavoidloads") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Avoidloads") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComStackInfo(fp);
        ComUseDefDag(fp);
        ComAliasAuxInfo(fp);

        flag |= OptAvoidLoads(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postavoidloads") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preavoidstores") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Avoidstores") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComStackInfo(fp);
        ComUseDefDagLoadStoreExtBbl(fp);
        ComAliasAuxInfo(fp);
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        
        flag |= OptAvoidStores(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postavoidstores") )
        PrintAll(stdout);
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/    

#if 0        
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preavoidstores") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Avoidstores") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComStackInfo(fp);
        ComUseDefDag(fp);
        ComAliasAuxInfo(fp);

        
        flag |= OptAvoidStores(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postavoidstores") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif

    /***************************************************/    

#if 0
     if( HasOption(PRINT_OPT_WHAT,"Infstat") )
         InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prestacksave") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Stacksave") )
    {
        ComputeSavedChangedRegisters(fp);
        ComputeLiveness(fp);
        flag |= OptStackSaveRegs(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Poststacksave") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prestackdel") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Stackdel") )
    {
        ComStackInfo(fp);
        flag |= OptDeleteStack(fp);
        flag |= OptKillDeadStores(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Poststackdel") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    return flag;
}


/*=======================================================
@FUNCTION
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeMoves"

static BOOL OptimizeMoves(FILE *fp)
{
    BOOL flag = 0;
    BOOL cont;
    
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preuseless") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Useless") )
    {
        do
        {
            cont = 0;
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            cont = OptKillUseless(fp);
            flag |= cont;
        }
        while( cont );
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postuseless") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    InfStatMove(fp);    

     if( HasOption(PRINT_OPT_WHAT,"Infstat") )
         InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Presignextensionone") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Signextensionone") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptSignExtensionOne(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postsignextensionone") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    /***************************************************/

    InfStatMove(fp);    

     if( HasOption(PRINT_OPT_WHAT,"Infstat") )
         InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Presignextensiontwo") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Signextensiontwo") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptSignExtensionTwo(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postsignextensiontwo") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/

    InfStatMove(fp);
        
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);

        
    if( HasOption(PRINT_OPT_WHEN,"Prenoop") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Noop") )
        flag |= OptEliminateNoops(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postnoop") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    InfStatMove(fp);
    
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Premove") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Move") )
    {
        do
        {
            cont = 0;
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            cont |= OptMoveElim(fp);
            cont |= OptMoveElimOne(fp);
            flag |= cont;
        }
        while( cont );
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postmove") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    
    /***************************************************/

    InfStatMove(fp);

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Premove") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Moveelimone") )
    {
        flag |= OptMoveElimOne(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postmove") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/
#if 1
    InfStatMove(fp);

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Precopy") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Copy") )
    {
        flag |= OptCopyProp(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcopy") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif
    return flag;
}

/*=======================================================
@FUNCTION
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeJumpTablesAndEscapes"

static BOOL OptimizeJumpTablesAndEscapes(FILE *fp)
{
    BOOL flag = 0;
    BOOL cont;

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preuseless") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Useless") )
    {
        do
        {
            cont = 0;
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            cont = OptKillUseless(fp);
             flag |= cont;
        }
        while( cont );
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postuseless") )
        PrintAll(stdout);

    /***************************************************/    
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

        if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preescapes") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Escapes") )
    {
        flag |= OptSpecialEliminateEscapes(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postescapes") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    
    /***************************************************/


    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preswitch") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Switch") )
        flag |= OptTableJumps(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postswitch") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Pregprel") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Gprel") )
        flag |= OptGprel(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postgprel") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    return flag;
}

/*=======================================================
@FUNCTION
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeConstants"

static BOOL OptimizeConstants(FILE *fp)
{
    BOOL flag = 0;
    BOOL cont;

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preuseless") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Useless") )
    {
        do
        {
            cont = 0;
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            cont = OptKillUseless(fp);
             flag |= cont;
        }
        while( cont );
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postuseless") )
        PrintAll(stdout);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preconstgenone") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Constgenone") )
        flag |= OptConstGenOne(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postconstgenone") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Preconstgentwo") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Constgentwo") )
    {
        flag |= OptConstGenTwo(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postconstgentwo") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Preconstuseone") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Constuseone") )
     {
         ComputeSavedChangedRegisters(fp);
         ComLiveness(fp);
         flag |= OptConstUseOne(fp);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postconstuseone") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);

     /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Preconstusetwo") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Constusetwo") )
     {
         ComputeSavedChangedRegisters(fp);
         ComLiveness(fp);
         flag |= OptConstUseTwo(fp);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postconstusetwo") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);


     /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Preconstusethree") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Constusethree") )
     {
         flag |= OptConstUseThree(fp);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postconstusethree") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);

     return flag;
}


/*=======================================================
@FUNCTION
Iteratively Execute Easy Optimization
This function calls all the easy optimizations.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeEasy"

extern BOOL OptimizeEasy(FILE *fp)    
{
    BOOL flag = 0;
    BOOL cont;

#if 0
    static BOOL x =0;

    if( x )
        return 0;
    else
        x = 1;
#endif

    InfoPhase(fp,"Optimize Simple");    
    
    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);

        
    if( HasOption(PRINT_OPT_WHEN,"Prenoop") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Noop") )
        flag |= OptEliminateNoops(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Postnoop") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);

    if( HasOption(PRINT_OPT_WHEN,"preReach") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Reach") )
    {
        flag |= OptRemoveUnreachableCode(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postReach") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    InfStatResource(fp);
        
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"preNormalize") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Normalize") )
    {
        flag |= OptNormalize(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postNormalize") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"preCalleesave") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Calleesave") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptCalleeSaveRegisters(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"postCalleesave") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preuseless") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Useless") )
    {
        do
        {
            cont = 0;
            ComputeSavedChangedRegisters(fp);
            ComLiveness(fp);
            cont = OptKillUseless(fp);
             flag |= cont;
        }
        while( cont );
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postuseless") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prepeep") )
        PrintAll(stdout);
        
    if( !HasOption(DISABLED_OPT,"Peep") )
        flag |= OptPeepHole(fp);

    if( HasOption(PRINT_OPT_WHEN,"Postpeep") )
    {
        /* needs liveness */
        PrintAll(stdout);
    }
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preeval") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Eval") )
        flag |= OptPartialEvaluate(fp);
       
    if( HasOption(PRINT_OPT_WHEN,"Posteval") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    InfStatResourceInsEval(fp);

    /***************************************************/

    /* this should go before reachabilty analysis to avoid
       some inconsistencies (fall thru convention) */

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preforward") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Forward") )
        flag |= OptBranchForwarding(fp);

    if( HasOption(PRINT_OPT_WHEN,"Postforward") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Precond") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Cond") )
    {
        flag |= OptBranchConditional(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcond") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Preempty") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Emptybbl") )
    {
        flag |= OptUnlinkEmptyBbls(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postempty") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Premerge") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Mergebbl") )
        flag |= OptBranchMerge(fp);

    if( HasOption(PRINT_OPT_WHEN,"Postmerge") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/

    flag |= OptimizeConstants(fp);

    /***************************************************/    
    
    flag |= OptimizeJumpTablesAndEscapes(fp);

    /***************************************************/    

    flag |= OptimizeMoves(fp);

    /***************************************************/    

    flag |= OptimizeLoadStores(fp);

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
        
    if( HasOption(PRINT_OPT_WHEN,"Prespec") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Spec") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComStackInfo(fp);
        ComUseDefDagLoadStoreExtBbl(fp);
        ComAliasAuxInfo(fp);
            
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);

        flag |= OptUndoSpeculations(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postspec") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);


    /***************************************************/

    /* this should happen before the conditional move stuff */

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Premergeinsup") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Mergeinsup") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComStackInfo(fp);
        ComUseDefDagLoadStoreExtBbl(fp);
        ComAliasAuxInfo(fp);
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptMergeInsUp(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postmergeinsup") )
        PrintAll(stdout);
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Precondpropagate") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Condpropagate") )
    {
        flag |= OptCondPropagate(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcondpropagate") )
        PrintAll(stdout);
    
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Precondmoveone") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Condmoveone") )
    {
        flag |= OptCondMoveOne(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcondmoveone") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);

    if( HasOption(PRINT_OPT_WHEN,"Precondmovetwo") )
        PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Condmovetwo") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptCondMoveTwo(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcondmovetwo") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    
#if 0
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Precompress") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Compress") )
    {
        OptNormalizeCond(fp);
        flag |= OptCompress(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postcompress") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Premergeinsdown") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Mergeinsdown") )
    {
        OptRemoveUnreachableCode(fp); /* necessary because AvoidLoads Aliases which use extended bbls */
        ComStackInfo(fp);
        ComUseDefDagLoadStoreExtBbl(fp);
        ComAliasAuxInfo(fp);        
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        flag |= OptMergeInsDown(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postmergeinsdown") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
    if( HasOption(PRINT_OPT_WHEN,"Predupbblone") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Dupbblone") )
    {
        flag |= OptDupBblOne(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postdupbblone") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Preinlinesmall") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Inlinesmall") )
     {
         flag |= OptInlineSmall(fp,NO_CODE_GROWTH);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postinlinesmall") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);
 
     /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
     
    if( HasOption(PRINT_OPT_WHEN,"Predupbbltwo") )
         PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Dupbbltwo") )
    {
        flag |= OptDupBblTwo(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postdupbbltwo") )
        PrintAll(stdout);
 
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
    
    /***************************************************/    
#if 0    
    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
     
    if( HasOption(PRINT_OPT_WHEN,"Predupbblthree") )
         PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Dupbblthree") )
    {
        flag |= OptDupBblThree(fp,NO_CODE_GROWTH);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Postdupbblthree") )
        PrintAll(stdout);
 
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);
#endif

    /***************************************************/

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
     
    if( HasOption(PRINT_OPT_WHEN,"Precommonsubexpresions") )
         PrintAll(stdout);
    
    if( !HasOption(DISABLED_OPT,"Commonsubexpressions") )
    {
        ComUseDefDagBbl(fp);
        flag |= OptCommonSubexpressions(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"Postcommonsubexpresions") )
        PrintAll(stdout);
 
    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    

     if( HasOption(PRINT_OPT_WHAT,"Infstat") )
         InfStatResource(fp);
         
     if( HasOption(PRINT_OPT_WHEN,"Preregreallocone") )
         PrintAll(stdout);
      
     if( !HasOption(DISABLED_OPT,"Regreallocone") )
     {
          ComputeSavedChangedRegisters(fp);
          ComLiveness(fp);
          OptRegisterReallocOne(fp);
      }
     
     if( HasOption(PRINT_OPT_WHEN,"Postregreallocone") )
         PrintAll(stdout);
     
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);
     
      /***************************************************/

     InfStatResource(fp);
     
     return flag;    
}


/*=======================================================
@FUNCTION
The Hard Optimization.
This function is called only once sandwiched between calls
to OptimizeEasy.
It calls all the optimizations that should be applied only once.
Because of problems with jumptable an ugly call to OptimizeEasy
was necessary.
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "OptimizeHard"


extern BOOL OptimizeHard(FILE *fp)    
{

    BOOL flag = 0;

    InfoPhase(fp,"Optimize Hard");    
    
    if( HasOption(DISABLED_OPT,"Hard") )
        return 0;

    ComputeSavedChangedRegisters(fp);
    ComLiveness(fp);
    ComStackInfo(fp);

#if 0
    OptPartialEvaluate(fp);
    OptRemoveUnreachableCode(fp);
    ComDomsAndLoops(fp);
    ComputeEdgeCounts(fp);
    
    ComUseDefDag(fp);
    ValProfInstrument(fp);
    
    /* OptDupRegionOne(fp, NO_CODE_GROWTH);    */
#endif

    /***************************************************/
    
    if( HasOption(PRINT_OPT_WHEN,"Preremovehelledges") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Removehelledges") && !HasOption(DISABLED_OPT,"Eval") )
    {
        OptRemoveHellEdges(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"Postremovehelledges") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/
    

    if( HasOption(PRINT_OPT_WHEN,"Precallingconvention") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Callingconvention") && !HasOption(DISABLED_OPT,"Eval") )
    {
        OptCallingConvention(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"Postcallingconvention") )
        PrintAll(stdout);

    if( HasOption(PRINT_OPT_WHAT,"Check") )
        CheckAll(fp);

    /***************************************************/    
#if 0
    if( HasOption(PRINT_OPT_WHEN,"Preshrinkwrap") )
        PrintAll(stdout);

    if( !HasOption(DISABLED_OPT,"Shrinkwrap") )
    {
        OptRemoveUnreachableCode(fp); 
        ComStackInfo(fp);
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComDomsAndLoops(fp);
        OptShrinkWrap(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"Postshrinkwrap") )
        PrintAll(stdout);
#endif
    
    /***************************************************/    
#if 0
    if( HasOption(PRINT_OPT_WHEN,"Predupfun") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Dupfun") )
    {
        OptRemoveUnreachableCode(fp);
        flag |= OptDuplicateFunctions(fp);
    }

    
    if( HasOption(PRINT_OPT_WHEN,"Postdupfun") )
        PrintAll(stdout);
#endif
    /***************************************************/
#if 1

    OptimizeEasy(fp);
                                 
#endif    
    /***************************************************/
        
#if 0
    IndirInit();

    /***************************************************/    

    if( !HasOption(DISABLED_OPT,"Indir") )
    {
        ComputeSavedChangedRegisters(fp);
        ComLiveness(fp);
        ComDomsAndLoops(fp);

	OptInlineIndir(fp);
    }


    if( HasOption(PRINT_OPT_WHEN,"PostIndir") ) 
        PrintAll(stdout);

    /***************************************************/    

    if( !HasOption(DISABLED_OPT,"Jump") )
    {
        ComputeNumberOps(fp);

        OptInlineJumps(fp);

        ComputeNumberOps(fp);
    }

    if( HasOption(PRINT_OPT_WHEN,"PostJump") ) 
        PrintAll(stdout);
    /***************************************************/

    IndirExit();
#endif

    if( HasOption(PRINT_OPT_WHEN,"Preinline") )
        PrintAll(stdout);

    if (!HasOption(DISABLED_OPT,"Inline") )
    {
	if( HasOption(INLINE_METHOD,"Crit") )
	{
	    ComputeCriticalBbls(fp);
	    ComStackInfo(fp);
	    OptInline(fp,0);
	}
	else if( HasOption(INLINE_METHOD,"Path") )
	{
	    ComputeCriticalBbls(fp);
	    ComStackInfo(fp);
	    OptInline(fp,1);
	}
    }

#if 0
    ComputeFunctionSideEffects(stdout);
    ComputeMod2kptrs();
#endif
    
    if( HasOption(PRINT_OPT_WHEN,"Postinline") )
        PrintAll(stdout);

    /***************************************************/    

    if( HasOption(PRINT_OPT_WHAT,"Infstat") )
        InfStatResource(fp);
    
     if( HasOption(PRINT_OPT_WHEN,"Preinlineonecall") )
         PrintAll(stdout);
       
     if( !HasOption(DISABLED_OPT,"Inlineonecall") )
     {
         flag |= OptInlineOneCall(fp);
     }
     
     if( HasOption(PRINT_OPT_WHEN,"Postinlineonecall") )
         PrintAll(stdout);
 
     if( HasOption(PRINT_OPT_WHAT,"Check") )
         CheckAll(fp);
 
     fflush(fp);
     
    /***************************************************/

    if( HasOption(PRINT_OPT_WHEN,"Prestackmerge") )
        PrintAll(stdout);
      
    if( !HasOption(DISABLED_OPT,"Stackmerge") )
    {
        ComStackInfo(fp);
        flag |= OptStackMerge(fp);
    }
    
    if( HasOption(PRINT_OPT_WHEN,"Poststackmerge") )
        PrintAll(stdout);

    /***************************************************/    


    return flag;
}


/*=======================================================
=======================================================*/


