/*=======================================================
=======================================================*/

#include  "global.h"


Indir *aIndir;
NUMBER aIndirFuns;
Indir *aValues;
NUMBER aValueCount;

/*=======================================================
=======================================================*/

BOOL flag_important_only =0;
BOOL flag_floats = 0;
double cut_off_percentile = 10.0;
INT64 cut_off_count = 0;

/*=======================================================
=======================================================*/

static int cmp(const void *a, const void *b)
{
    indir *v1 = (indir *)a;
    indir *v2 = (indir *)b;

    return (v2->count - v1->count);
}


/*=======================================================
=======================================================*/

extern INT64 BinaryLog(INT64 n)
{
    INT64 t=0;
    
    if( n <= 0 ) return t;

    if( n >> 32 ) { n >>= 32; t += 32;}
    if( n >> 16 ) { n >>= 16; t += 16;}
    if( n >>  8 ) { n >>=  8; t +=  8;}
    if( n >>  4 ) { n >>=  4; t +=  4;}
    if( n >>  2 ) { n >>=  2; t +=  2;}
    if( n >>  1 ) { n >>=  1; t +=  1;}

    return t;
}

/*=======================================================
 we want to have the value profile create an ascii file
 so we can look at the values used 
=======================================================*/

extern void ReadProfilingInformation(char *filename)
{
    FILE *fp;
    INT64 cnt;
    INT64 bcount = 0;
    INT64 ecount = 0;
    INT64 fcount = 0;
    INT64 mapcount = 0;
    INT64 totalcount;
    int i;
    int j;
    int temp;
    INT64 counter;


    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    fp = fopen(filename,"r");
    if( fp == 0 )
    {
        
        fprintf(stderr,"Cannot open adr file: %s, no value profiles\n",filename);
        return;
    }
    
    fread(&cnt,sizeof(cnt),1,fp);


    if ( cnt != ALTO_MAGIC )
    {
	return;
    }


    fread(&bcount,sizeof(bcount),1,fp);
    fread(&ecount,sizeof(ecount),1,fp);
    fread(&fcount,sizeof(fcount),1,fp);
    fread(&mapcount,sizeof(mapcount),1,fp);

    printf("######################################################################\n");
    printf(" HEADER: %s\n",filename);
    printf("######################################################################\n");
    
    printf("using alto profiling, checking counters\n");
    printf("number of bbl counters is %ld\n", bcount);
    printf("number of edg counters is %ld\n", ecount);
    printf("number of fun counters is %ld\n", fcount);
    printf("number of iins map is %ld\n", mapcount);


    if( bcount > 0 )
    {
        INT64 count[64];
        INT64 hi = 0;
        INT64 lo = 0;

        memset( count,0,sizeof(count));
    
        printf("######################################################################\n");
        printf("BBL COUNTS\n");
        printf("######################################################################\n");


        
        for (i=0; i < bcount; i++)
        {
            fread(&counter, sizeof(counter), 1, fp);
            /* printf("%6d %10ld\n",i,counter); */
            if( counter > hi ) hi = counter;
            if( counter < lo ) lo = counter;
            count[BinaryLog(counter)]++;
        }

        printf("======================================================================\n");
        printf("log      n            bbls\n");
        printf("======================================================================\n");

        for (i=0; i < 64; i++)
        {
            if( count[i] == 0 ) continue;
            printf("%2d  %13ld   %5ld\n",i,1L<<i,count[i]);
        }
        printf("======================================================================\n");
        printf("hi: %ld  lo: %ld\n",hi,lo);
    }
    
    if( ecount > 0 )
    {
        printf("######################################################################\n");
        printf(" EDG COUNTS\n");
        printf("######################################################################\n");
        
        for (i=0; i < ecount; i++)
        {
            INDEX iedg;
            
	    fread(&iedg, sizeof(counter), 1, fp);
	    fread(&counter, sizeof(counter), 1, fp);
        }
    }
    
    if( fcount > 0 )
    {
        printf("######################################################################\n");
        printf("VALUE PROFILING\n");
        printf("######################################################################\n");
        
        aValues = malloc(sizeof(Indir) * fcount);
        aValueCount = fcount;
        
        for (i=0; i < fcount; i++)
        {
            fread(&counter, sizeof(counter), 1, fp);
            /* bottom half of the word is the instruction, top is register */
            aValues[i].iins = counter;
            
            aValues[i].reg1 = (counter >> 32) & 0xffff;
            aValues[i].reg2 = (counter >> 48) - 1;
            aValues[i].ins=  counter & 0xffffffff;
            
            fread(&counter, sizeof(counter), 1, fp);
            
            totalcount = 0;
            for (j=0; j < CONFIG_NINDIR; j++)
            {
                fread(&counter, sizeof(counter), 1, fp);
                aValues[i].values[j].addr = counter;
                fread(&counter, sizeof(counter), 1, fp);
                aValues[i].values[j].count = counter;
                totalcount += counter;
            }
            aValues[i].totalcount = totalcount;
            
            qsort(aValues[i].values, CONFIG_NINDIR - 1, sizeof(indir), &cmp);
        }
        
        for (temp=0; temp < fcount; temp++)
        {
            if( aValues[temp].totalcount < cut_off_count ) continue;
            
            printf("======================================================================\n");
            printf("id: %20ld=0x%016lx  cnt: %ld  reg1: %2d  reg2 %2d\n", 
                   aValues[temp].iins, aValues[temp].iins, aValues[temp].totalcount,aValues[temp].reg1, aValues[temp].reg2);
            if( flag_important_only )
                if( 100.0*aValues[temp].values[0].count/(aValues[temp].totalcount+1) < cut_off_percentile )
                {
                printf("UNINTERRSTING\n");
                continue;
                }
            
            for (j=0; j < CONFIG_NINDIR; j++)
            {
                if( aValues[temp].values[j].count == 0 ) continue;
                printf("val: %16lx  (%f) cnt: %20ld (%4.1f%%)\n",
                       aValues[temp].values[j].addr, *(double*) &aValues[temp].values[j].addr, aValues[temp].values[j].count,
                       100.0*aValues[temp].values[j].count/(aValues[temp].totalcount+1));
            }
        }
        printf("======================================================================\n");
    }
    
    fclose(fp);
    
}

/*=======================================================
=======================================================*/

static void Usage()
{
    printf("usage: valprofdump  [-i] value-profiling-file\n"
           "\n"
           "Prints the contents of value-profiling-file\n"
           "\n"
           "-c num\tonly print items with totalcount > c\n"
           "-i\tonly print important entries\n");
}

/*=======================================================
=======================================================*/

extern int main(int argc, char *argv[])
{

    int i;
    
    while( EOF != (i = getopt(argc,argv,"ihc:")) )
    {
        switch(i)
        {
        case 'c':
            cut_off_count = atoi(optarg);
            break;
        case 'i':
            flag_important_only = 1;
            break;
        case 'f':
            flag_floats = 1;
            break;
        case 'h':
        case '?':
        default:
            Usage();
            break;
        }
    }

    argv += optind;     
    argc -=optind;

    ReadProfilingInformation(argv[0]);
    return 0;
}


/*=======================================================
=======================================================*/





