#include "global.h"
#include "functions.h"
#include "flags.h"

#ifdef SYS_COFF
#include <stdio.h>
#include <cmplrs/prof_header.h>
#include <cmplrs/pdf.h>
#include <stdlib.h>
#endif

#define FILE_PDF 0x0a213e4644503c21L
#define FILE_ADR 0x0f0e0010L
#define FILE_CNT 0x0f0e0a11L
#define FILE_CNT2 0x0f0e0a33L

#define USE_NEW_ATOM_PIXIE 1
 
 
 
Indir *aIndir;
NUMBER aIndirFuns;
Indir *aValues;
NUMBER aValueCount;

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ProfReadPixieCounts"

extern BOOL ProfReadPixieCounts(FILE *fp)
{
    INDEX ibbl;
    FILE *fp_adr;
    FILE *fp_cnt;
    char fn_adr[256];
    char fn_cnt[256];
    int  adr;
    INT64 cnt;
    NUMBER rcount = 0;
    NUMBER gcount = 0;
    
    /* otherwise, use the pixie data */

    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    sprintf(fn_adr,"%s.Addrs",CONFIG_INFILE);

    fp_adr = fopen(fn_adr,"r");
    if( fp_adr==0 )
    {
        return 0;
    }
    
    fread(&cnt,sizeof(cnt),1,fp_adr);

    if( (cnt & 0xffffffff) == FILE_ADR )
    {
        FATAL(("Your are using and old pixie version\n"));
    }
    else if( cnt == FILE_PDF  )
    {
#if USE_NEW_ATOM_PIXIE
        fseek(fp_adr,0x100,SEEK_SET);
        /* read string length */
        fread(&cnt,sizeof(cnt),1,fp_adr);
        /* round up to multiple of 4*/
        cnt += 3;
        cnt &= ~3;

        /* skip next int32 */
        cnt += 4;
        
        /* round up to multiple of 8 */
        cnt += 7;
        cnt &= ~7;
        
        fseek(fp_adr,cnt,SEEK_CUR);
        fread(&cnt,sizeof(cnt),1,fp_adr);
        /*printf("adr (%ld),data offset: %lx\n",cnt,ftell(fp_adr));*/
#else
        FATAL(("You are using the new pixie\n"));
#endif
    }
    else
    {
        FATAL(("bad count file\n"));
    }   

    /*
      open count file and advance file pointer
      to beginning of count data
      */

    sprintf(fn_cnt,"%s.Counts",CONFIG_INFILE);
    
    fp_cnt = fopen(fn_cnt,"r");
    if( fp_cnt==0 )
    {
        return 0;
    }
    
    fread(&cnt,sizeof(cnt),1,fp_cnt);
    
    if( (cnt & 0xffffffff) == FILE_CNT )
    {
        FATAL(("Your are using and old pixie version\n"));
    }
    else if( (cnt & 0xffffffff) == FILE_CNT2 )
    {
        FATAL(("Your are using and old pixie version\n"));
    }
    else if( cnt == FILE_PDF  )
    {
#ifdef USE_NEW_ATOM_PIXIE
        fseek(fp_cnt,0x148,SEEK_SET);
        fread(&cnt,sizeof(cnt),1,fp_cnt);
        cnt += 3;
        cnt &= ~3;

        cnt += 4;
        cnt += 7;
        cnt &= ~7;

        fseek(fp_cnt,cnt,SEEK_CUR);
        
        fread(&cnt,sizeof(cnt),1,fp_cnt);

        /*        printf("cnt (%ld) data offset: %lx\n",cnt,ftell(fp_cnt));*/
#else
        FATAL(("You are using the new pixie\n"));
#endif
        
    }
    else
    {
        FATAL(("bad count file\n"));
    }
    
    
    for(;;)
    {
        if( 1 != fread(&cnt,sizeof(cnt),1,fp_cnt) )
            break;
        fread(&adr,sizeof(adr),1,fp_adr);


        rcount++;

        ibbl = AINS_BBL(adr);
#if 0
        printf("ins: %d  (%lx) bbl: %d cnt: %ld\n",adr,Ins2Addr(adr),ibbl,cnt);
#endif
        
               

#if 0
        if( AINS_FLAGS(adr) & FI_BBL_BEG )
        {
            ABBL_COUNT(ibbl) = cnt;
        }



        printf("iins: %d ibbl: %d \n",adr,ibbl);
        if( ibbl == 5441 )
        {
            printf("*****\n");
            
        }
#endif
        
#if 1
        if( VALID(ABBL_COUNT(ibbl)) )
        {
            if( !ABBL_IS_FIRST(ibbl) && !ABBL_IS_KILLED(ibbl) )
            {
                WARNING(("duplicate ins %d bbl %d  changing counts: %ld to %ld\n",
                         adr,ibbl,ABBL_COUNT(ibbl),cnt));
            }
            
        }
#endif

        ABBL_COUNT(ibbl) = cnt;
    }
    
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
        if( VALID(ABBL_COUNT(ibbl)) ) continue;
        if( ABBL_TYPE(ibbl)==BT_EXIT ) continue;

        if( ABBL_TYPE(ibbl-1) == BT_PAL )
        {
            ABBL_COUNT(ibbl) = ABBL_COUNT(ibbl-1);
            continue;
        }
        
#if 1
        WARNING(("guessing count %ld for bbl %d in fun %d [%s]\n",ABBL_COUNT(ibbl),ibbl,ABBL_FUN(ibbl),AFUN_NAME(ABBL_FUN(ibbl))));
#endif
        
        gcount++;
        
        ABBL_COUNT(ibbl) = 0; 
    }

    InfoStatus(fp,"\t[ProfReadPixieCounts] File %s  Read %d  Estimated: %d\n",fn_cnt,rcount,gcount);
    return 1;
}

/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ProfCheckCounts"

extern BOOL ProfCheckCounts(FILE *fp)
{
    INDEX ibbl;
    INDEX iedg;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        if( AFUN_IS_PSEUDO(ABBL_FUN(ibbl)) ) continue;
        if( ABBL_TYPE(ibbl)==BT_EXIT ) continue;

        /* if there is only one predecessor
           this bbl must have a less or equal count
           */
           
        if( AbblNpreds(ibbl) == 1)
        {
            INDEX hd = AEDG_HEAD( ABBL_PRED(ibbl) );

            if( AFUN_IS_PSEUDO(ABBL_FUN(hd)) ) continue;
            
            if( ABBL_COUNT(ibbl) > ABBL_COUNT(hd) )
            {
                WARNING(("inconsistency in bbl %d->%d  count: %ld -> %ld\n",
                       hd,ibbl,ABBL_COUNT(hd),ABBL_COUNT(ibbl)));
#if 0
                PrintBblSuper(stdout,hd);
                PrintBblSuper(stdout,ibbl);
#endif                
            }
            
        }

        /* if there is only one successor
           this bbl must have a less or equal count
           */
        
        if( AbblNsuccs(ibbl) == 1)
        {
            INDEX tl = AEDG_TAIL( ABBL_SUCC(ibbl) );

            if( AFUN_IS_PSEUDO(ABBL_FUN(tl)) ) continue;
            if( ABBL_TYPE(tl) == BT_EXIT ) continue;
            
            if( ABBL_COUNT(ibbl) > ABBL_COUNT(tl) )
            {
                WARNING(("inconsistency in bbl %d<-%d  count: %ld <- %ld\n",
                       tl,ibbl,ABBL_COUNT(tl),ABBL_COUNT(ibbl)));
#if 0
                PrintBblSuper(stdout,tl);
                PrintBblSuper(stdout,ibbl);
#endif                
            }
            
        }
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            INT64 mincount = 0;
        
            switch( AEDG_TYPE(iedg) )
            {
            case ET_NORM:
            case ET_FUN_CALL:
            case ET_FUN_LINK:
                mincount += ABBL_COUNT( AEDG_HEAD(iedg) );
                break;
            }

            if( ABBL_COUNT(ibbl) < mincount )
            {
                WARNING(("inconsistency in bbl %d  is: %ld  min: %ld\n",ibbl,ABBL_COUNT(ibbl),mincount));
#if 0
                PrintFunSuper(stdout,ABBL_FUN(ibbl));
#endif                
            }
        }
    }

    
    InfoStatus(fp,"\t[ProfCheckCounts]\n");

    return 1;
}

/*=======================================================
=======================================================*/

static int cmp(const void *a, const void *b)
{
    indir *v1 = (indir *)a;
    indir *v2 = (indir *)b;

    return (v2->count - v1->count);
}


extern BOOL ProfReadAltoCounts(FILE *fp)
{
    INDEX iedg;
    FILE *fp_adr;
    char fn_adr[256];
    INT64 cnt;
    INT64 bcount = 0;
    INT64 ecount = 0;
    INT64 fcount = 0;
    INT64 mapcount = 0;
    INT64 totalcount;
    int i;
    int j;
    int k;
    INT64 counter;
    INT64 initial;
    INT64 post;
    int verbose = 0;

    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    sprintf(fn_adr,"%s.AltoCounts",CONFIG_INFILE);
    
    fp_adr = fopen(fn_adr,"r");
    if(fp_adr==0)
    {
        return 0;
    }
    
    fread(&cnt,sizeof(cnt),1,fp_adr);


    if ( cnt == ALTO_MAGIC )
    {
	fread(&bcount,sizeof(bcount),1,fp_adr);
	fread(&ecount,sizeof(ecount),1,fp_adr);
	fread(&fcount,sizeof(fcount),1,fp_adr);
	fread(&mapcount,sizeof(mapcount),1,fp_adr);

	if (verbose)
        {
	    printf("using alto profiling, checking counters\n");
	    printf("number of bbl counters is %ld\n", bcount);
	    printf("number of edg counters is %ld\n", ecount);
	    printf("number of fun counters is %ld\n", fcount);
	    printf("number of iins map is %ld\n", mapcount);
	    printf("number of bbls is %d\n", CONFIG_UBBL);
	}
	
	
        for (i=0; i < bcount; i++)
        {
            fread(&counter, sizeof(counter), 1, fp_adr);
            /* don't change exit blocks */
            if (ABBL_IS_LAST(i)) continue;
            if ((counter != ABBL(i).Count) &&
                (ABBL_TYPE(i) != BT_EXIT ) &&
                (ABBL(i).Count >= 0) &&
                (!ABBL_IS_KILLED(i)) &&
                (AbblNInss(i) > 0) && 
                (abs(ABBL(i).Count - counter) > 1))
            {
                printf("something odd bbl %d counter %ld count of %ld\n", i, counter, ABBL(i).Count);
                PrintBblSuper(stdout, i);
            }
#if 0
            printf("count for bbl %d %lx is %ld\n", i, Ins2Addr(ABBL_INS_FIRST(i)), counter);
#endif
            ABBL(i).Count = counter;
	}
	
	fflush(stdout);
	for (i=0; i < ecount; i++) {
	    fread(&iedg, sizeof(counter), 1, fp_adr);
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    AEDG_COUNT(iedg) = counter;
	}

	aIndir = malloc(sizeof(Indir) * fcount);
	aIndirFuns = fcount;

	for (i=0; i < fcount; i++) {
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    aIndir[i].iins = Addr2Ins(counter);

	    if (verbose) {
		printf("reading in counts of instruction %ld\n", aIndir[i].iins);
		printf("address of instruction is %lx\n", counter);
	    }
	    fread(&counter, sizeof(counter), 1, fp_adr);

	    if (verbose)
		printf("current counter %ld\n", counter);

	    totalcount = 0;
	    for (j=0; j < CONFIG_NINDIR; j++) {
		fread(&counter, sizeof(counter), 1, fp_adr);
		aIndir[i].values[j].addr = counter;
		fread(&counter, sizeof(counter), 1, fp_adr);
		aIndir[i].values[j].count = counter;
		totalcount += counter;
	    }
	    aIndir[i].totalcount = totalcount;
	    qsort(aIndir[i].values, CONFIG_NINDIR, sizeof(indir), &cmp);
	    if (verbose) {
		printf("indir index is %d\n", i);
		for (j=0; j < CONFIG_NINDIR; j++) {
		    printf("address %lx has count of %ld\n", aIndir[i].values[j].addr, aIndir[i].values[j].count);
		}
	    }
	}

	/* the really clever thing to do here would be modify the address
	   in the aIndir structure based on the way this map goes */
	if (verbose)
	    printf("reading in instruction map\n");
	for (i=0; i < mapcount; i++) {
	    fread(&initial, sizeof(initial), 1, fp_adr);
	    fread(&post, sizeof(post), 1, fp_adr);
	    
	    for (j = 0; j < fcount; j++) {
		for (k = 0; k < CONFIG_NINDIR; k++) {
		    if (aIndir[j].values[k].addr == post) {
			aIndir[j].values[k].addr = initial;
		    }
		}
	    }
	    
	}

/* 	ComVerifyCounts(fp); */
	
    }

    InfoStatus(fp,"\t[ProfReadAltoCounts] File %s\n",fn_adr);
    return 1;
}

/*=======================================================
=======================================================*/

extern void ComVerifyCounts(FILE *fp) 
{
    INDEX ibbl;
    NUMBER tcount, fcount;

    FOREACH_BBL(ibbl) {
	if (ABBL_TYPE(ibbl) == BT_BRANCH) {
	    /* check to see if true and false edges add up */
	    tcount = AEDG_COUNT(ABBL_EDG_TRUE(ibbl));
	    fcount = AEDG_COUNT(ABBL_EDG_FALSE(ibbl));
	    
	    CERROR(((tcount + fcount) != ABBL_COUNT(ibbl)), ("ERROR true and false counts don't match bbl in bbl %d count!!!\n", ibbl));
	}
    }
}


/*=======================================================
 we want to have the value profile create an ascii file
 so we can look at the values used 
=======================================================*/
#undef __PROC__
#define __PROC__ "ComReadValueProfilingInformation"


extern void ComReadValueProfilingInformation(FILE *fp)
{
    FILE *fp_adr;
    char fn_adr[256];
    INT64 cnt;
    INT64 bcount = 0;
    INT64 ecount = 0;
    INT64 fcount = 0;
    INT64 mapcount = 0;
    INT64 totalcount;
    int i;
    int j;
    int temp;
    int last_print;
    INDEX prev_fun;
    INT64 counter;
#if 0
    int k;
    INT64 initial;
    INT64 post;
    double percentage;
#endif
    BOOL   verbose = HasOption(PRINT_OPT_VERBOSE,"Valprof");


    CASSERT( CONFIG_HAVE_BBL_PROFILE,("need bbl counts for this optimization\n"));

    /*
      open address file and advance file pointer
      to beginning of address data
      */
    
    sprintf(fn_adr,"%s.Value",CONFIG_INFILE);
    
    fp_adr = fopen(fn_adr,"r");
    if( fp_adr==0 )
    {
        
        WARNING(("Cannot open adr file: %s, no value profiles\n",fn_adr));
        return;
    }
    
    fread(&cnt,sizeof(cnt),1,fp_adr);


    if ( cnt != ALTO_MAGIC ) {
	return;
    }


    fread(&bcount,sizeof(bcount),1,fp_adr);
    fread(&ecount,sizeof(ecount),1,fp_adr);
    fread(&fcount,sizeof(fcount),1,fp_adr);
    fread(&mapcount,sizeof(mapcount),1,fp_adr);
    
    if (verbose) {
	printf("using alto profiling, checking counters\n");
	printf("number of bbl counters is %ld\n", bcount);
	printf("number of edg counters is %ld\n", ecount);
	printf("number of fun counters is %ld\n", fcount);
	printf("number of iins map is %ld\n", mapcount);
	printf("number of bbls is %d\n", CONFIG_UBBL);
    }
    
    if (bcount != 0) {
	ERROR(("bbl count is not zero in value profile\n"));
    }
    
    if (ecount != 0) {
	ERROR(("Edg count is not zero in value profile\n"));
    }
    
    aValues = malloc(sizeof(Indir) * fcount);
    CONFIG_NVALUES = fcount;


    FOREACH_VPF(i)
    {
	fread(&counter, sizeof(counter), 1, fp_adr);
	/* bottom half of the word is the instruction, top is register */
	aValues[i].iins = counter;
	AVPF_MAGIC(i) = counter;

	aValues[i].reg1 = (counter >> 32) & 0xffff;
	aValues[i].reg2 = (counter >> 48) - 1;
	aValues[i].ins=  counter & 0xffffffff;

	fread(&counter, sizeof(counter), 1, fp_adr);
	
	totalcount = 0;
	for (j=0; j < CONFIG_NINDIR; j++) {
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    aValues[i].values[j].addr = counter;
	    fread(&counter, sizeof(counter), 1, fp_adr);
	    aValues[i].values[j].count = counter;
	    totalcount += counter;
	}
	aValues[i].totalcount = totalcount;

	qsort(aValues[i].values, CONFIG_NINDIR - 1, sizeof(indir), &cmp);
	fflush(fp);

#if 0
	if (aValues[i].totalcount != 0) 
	    percentage = (double)aValues[i].values[0].count/(double)aValues[i].totalcount;
	else
	    percentage = 0.0;
	fprintf(valprof, "%8d %8d %8d %16lx %16ld %16f\n", 
		aValues[i].ins, aValues[i].reg1, aValues[i].reg2,
		aValues[i].values[0].addr,
		aValues[i].values[0].count,
		percentage);
#endif
    }
    
    if (verbose) {
	last_print = 0;
	
	for (i=0; i < fcount; i++) {
	    INDEX ifun = ABBL_FUN(AINS_BBL(aValues[i].iins & 0xffffffff));
	    
	    prev_fun = ABBL_FUN(AINS_BBL(aValues[last_print].iins & 0xffffffff));
	    
	    if ((ifun != prev_fun) || (i == fcount -1)) {
		if ((ifun == prev_fun) && (i == fcount-1)) i++;
		printf("MagicString %s%d\n", CONFIG_INFILE, i);
		for (temp = last_print; temp < i; temp++) {
		    printf("registers profiled %d and %d\n", aValues[temp].reg1, aValues[temp].reg2);
		    printf("reading in counts of instruction %d, has count of %ld\n", aValues[temp].ins , aValues[temp].totalcount);
		    fflush(stdout);
		    for (j=0; j < CONFIG_NINDIR; j++) {
			printf("address %16lx has count of %20ld\n", aValues[temp].values[j].addr, aValues[temp].values[j].count);
			fflush(stdout);
		    }
		}
		PrintFunSuper(stdout, prev_fun);
		printf("MagicString %s%d\n", CONFIG_INFILE, i);
		fflush(stdout);
		last_print = i;
		if (i == fcount-1) {
		    printf("MagicString %s%d\n", CONFIG_INFILE, i);
		    for (temp = last_print; temp < i; temp++) {
			printf("registers profiled %d and %d\n", aValues[temp].reg1, aValues[temp].reg2);
			printf("reading in counts of instruction %d, has count of %ld\n", aValues[temp].ins , aValues[temp].totalcount);
			fflush(stdout);
			for (j=0; j < CONFIG_NINDIR; j++) {
			    printf("address %16lx has count of %20ld\n", aValues[temp].values[j].addr, aValues[temp].values[j].count);
			    fflush(stdout);
			}
		    }
		    PrintFunSuper(stdout, prev_fun);
		    printf("MagicString %s%d\n", CONFIG_INFILE, i);
		    fflush(stdout);
		}
	    }
	}
    }
    

#if 0
    /* the really clever thing to do here would be modify the address
       in the Values structure based on the way this map goes */
    if (verbose)
	printf("reading in instruction map\n");
    for (i=0; i < mapcount; i++) {
	fread(&initial, sizeof(initial), 1, fp_adr);
	fread(&post, sizeof(post), 1, fp_adr);
	
	for (j = 0; j < fcount; j++) {
	    for (k = 0; k < CONFIG_NINDIR; k++) {
		if (aValues[j].values[k].addr == post) {
		    aValues[j].values[k].addr = initial;
		}
	    }
	}
    }
#endif

    fclose(fp_adr);
    InfoStatus(fp,"\t[PrfValues] %d profiles read from %s\n",fcount,fn_adr);
}


/*=======================================================
=======================================================*/





