/*=======================================================
  Author:   R. Muth
  Version:  $Id: usedefdag.c,v 1.4 1999/06/07 03:51:21 muth Exp $
  Comment:  Single UseDef Chain Creation
=======================================================*/


#include "global.h" 
#include "functions.h"
#include "opcode.h"
#include "regs.h"
#include "flags.h"

#include <string.h>

/*=======================================================
=======================================================*/

#define ABBL_USEDEF_IN(i)  ABBL_U1(i).doubleword.d1
#define ABBL_USEDEF_DEF(i) ABBL_U1(i).doubleword.d2
#define ABBL_USEDEF_CONFLUENCE(i) ABBL_U2(i).quadword

#define USEDEF_BOT -69
#define USEDEF_TOP -1

/*=======================================================
=======================================================*/
/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "UseDefComputeJoin"

static void UseDefComputeJoin(INDEX ibbl, INDEX def)
{
    CERROR( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63,("weird def %x propagated to bbl %d\n",def,ibbl));
        
    if( def == ABBL_USEDEF_IN(ibbl) ) return;    
    if( def == USEDEF_BOT ) return;
    if( ABBL_USEDEF_IN(ibbl) == USEDEF_TOP ) return;

    ABBL_MARK(ibbl);
    
    if( def == USEDEF_TOP )
    {
        ABBL_USEDEF_IN(ibbl) = def;
        return;
    }

    if( ABBL_USEDEF_IN(ibbl) == USEDEF_BOT )
    {
        ABBL_USEDEF_IN(ibbl) = def;
        return;
    }

    ABBL_USEDEF_IN(ibbl) = USEDEF_TOP;
}

/*=======================================================
=======================================================*/

#define CFA1

#undef __PROC__
#define __PROC__ "UseDefPropagation"

static void UseDefPropagation(REG reg)
{
    BOOL  flag;
    INDEX iins;
    INDEX ibbl;
    INDEX ifun;
    
    INDEX iedg;
    INDEX def;
        
    /* Initialize Fixpoint Computation */

    FOREACH_LIVE_BBL(ibbl)
    {
        ifun = ABBL_FUN(ibbl);
        
        if( AFUN_IS_PSEUDO(ifun) )
        {
            ABBL_USEDEF_IN(ibbl) = USEDEF_TOP;
            ABBL_USEDEF_DEF(ibbl) = USEDEF_TOP;
        }
        else if( AbblNpreds(ibbl) == 0 )
        {
            ABBL_USEDEF_IN(ibbl) = USEDEF_TOP;
            ABBL_USEDEF_DEF(ibbl) = USEDEF_BOT;
        }
        else
        {
            ABBL_USEDEF_IN(ibbl)  = USEDEF_BOT;
                            
            if( ABBL_USEDEF_CONFLUENCE(ibbl) & REG2BIT(reg) )
            {
                ABBL_USEDEF_DEF(ibbl) =   CONFIG_NINS + (ibbl * 64) + reg;
            }
            else
            {
                ABBL_USEDEF_DEF(ibbl)  = USEDEF_BOT;
            }
            
            FOREACH_BBL_INS_R(ibbl,iins)
            {
                if( AINS_REG_C(iins) == reg )
                {
                    ABBL_USEDEF_DEF(ibbl) = iins;
                    break;
                }
            }
        }

        def =  ABBL_USEDEF_DEF(ibbl);
        CERROR( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63,("reg: %d weird def %x propagated to bbl %d\n",reg,def,ibbl));

        def = ABBL_USEDEF_IN(ibbl);
        CERROR( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63,("reg: %d weird def %x propagated to bbl %d\n",reg,def,ibbl));
        
        ABBL_MARK(ibbl);
    }

    do
    {
        flag = 0;

        /* printf("ROUND: %d\n",reg); */
        
        FOREACH_LIVE_BBL(ibbl)
        {
            
            if( !ABBL_IS_MARKED(ibbl) ) continue;
            
            ABBL_UNMARK(ibbl);
            flag = 1;


            /* COMPUTE OUT SET */

            def = ABBL_USEDEF_DEF(ibbl);
            if( def == USEDEF_BOT )
                def = ABBL_USEDEF_IN(ibbl);

            if( def == USEDEF_BOT ) continue;
            
            CERROR( def > CONFIG_NINS + (CONFIG_UBBL * 64) + 63,("reg: %d weird def %x propagated to bbl %d\n",reg,def,ibbl));
    
            /* PROPAGATE OUT SET TO INSETS */

            FOREACH_BBL_SUCC(ibbl,iedg)
            {
                INDEX tl = AEDG_TAIL(iedg);

#ifdef CFA1
                /* propagate unchanged regs */
                   
                if( AEDG_TYPE(iedg) == ET_FUN_LINK )
                {
                    INDEX CallFun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
                    REGLIST NotChanged = ~AFUN_REGS_CHANGED(CallFun);

                    if( REG2BIT(reg) & NotChanged )
                    {
                        UseDefComputeJoin(tl,def);
                    }
                    continue;
                }
  
                /* propagate changed regs only */
                
                if( AEDG_TYPE(iedg) == ET_FUN_RETURN )
                {
                    INDEX CallFun = ABBL_FUN(ibbl);
                    REGLIST Changed = AFUN_REGS_CHANGED(CallFun);
                    if( REG2BIT(reg) & Changed )
                    {
                        UseDefComputeJoin(tl,def);
                    }
                    
                    continue;
                }
#else
                if( AEDG_TYPE(iedg) == ET_FUN_LINK )
                {
                    continue;
                }
                
#endif
                UseDefComputeJoin(tl,def);
            }
            
        }
    }
    while( flag );
}


/*=======================================================
=======================================================*/

#undef __PROC__
#define __PROC__ "UseDefPropagateDefsToInstructions"

static void UseDefPropagateDefsToInstructions(REG reg)
{
    INDEX ifun;
    INDEX iins;
    INDEX ibbl;
    INDEX def;

    FOREACH_LIVE_FUN(ifun)
    {
        FOREACH_FUN_BBL(ifun,ibbl)
        {
            def = ABBL_USEDEF_IN(ibbl);
            CERROR( def == USEDEF_BOT,("unexpected bottom reg: %d\n",reg));
            if( def == USEDEF_TOP )
            {
                CASSERT( ABBL_USEDEF_CONFLUENCE(ibbl) & REG2BIT(reg),("top without confluence point  reg: %d\n",reg));
            }

            if( ABBL_USEDEF_CONFLUENCE(ibbl) & REG2BIT(reg) )
            {
                def = CONFIG_NINS + (ibbl * 64) + reg;
            }
        
            FOREACH_BBL_INS(ibbl,iins)
            {
                if( AINS_REG_A(iins) == reg && !(AINS_FLAGS(iins) & FI_EVAL_A) )
                    AINS_A_EVAL(iins) = def;
                
                if( AINS_REG_B(iins) == reg && !(AINS_FLAGS(iins) & FI_EVAL_B) )
                    AINS_B_EVAL(iins) = def;
                
                if( AINS_REG_C(iins) == reg  )
                    def = iins;
            }
        }
    }    
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "UseDefInitialConfluencePoints"

static void UseDefInitialConfluencePoints()
{
    INDEX ifun;
    INDEX ibbl;
    INDEX jbbl;
    INDEX pred;
    INDEX iedg;
    
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_USEDEF_CONFLUENCE(ibbl) = REGS_NONE;
    }
    
    /* all function init blocks are confluence points */
    FOREACH_LIVE_FUN(ifun)
    {
        ibbl = AFUN_BBL_FIRST(ifun);
        ABBL_USEDEF_CONFLUENCE(ibbl) = REGS_ALL_BUT_ZERO;
    }

    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_TYPE(ibbl) != BT_CALL && ABBL_TYPE(ibbl) != BT_PAL ) continue;
        ifun = ABBL_FUN(AEDG_TAIL(ABBL_EDG_CALL(ibbl)));
        jbbl = AEDG_TAIL(ABBL_EDG_LINK(ibbl));
        ABBL_USEDEF_CONFLUENCE(jbbl) = AFUN_REGS_CHANGED(ifun);
    }

    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            pred = AEDG_HEAD(iedg);
            
            /* this is a quick hack --- NEEDS WORK */
            
            if( AEDG_TYPE(iedg) == ET_HELL || AEDG_TYPE(iedg) == ET_COMPENSATE  )
            {
                ABBL_USEDEF_CONFLUENCE(ibbl) = REGS_ALL_BUT_ZERO;
                break;
            }
        }
        
        if( AbblNpreds(ibbl) == 0 )
        {
            ABBL_USEDEF_CONFLUENCE(ibbl) = REGS_ALL_BUT_ZERO;
            continue;
        }
    }    
}



/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "UseDefMoreConfluencePoints"

static NUMBER UseDefMoreConfluencePoints(REG reg)
{
    INDEX ibbl;
    INDEX iedg;
    INDEX pred;
    INDEX def;
    
    BOOL flag=0;

    NUMBER count=0;
    NUMBER pot = 0;


    FOREACH_LIVE_BBL(ibbl)
    {
        if( ABBL_USEDEF_CONFLUENCE(ibbl) & REG2BIT(reg) ) continue;
        if( ABBL_USEDEF_IN(ibbl) != USEDEF_TOP ) continue;

        flag = 1;

        pot++;
        
        FOREACH_BBL_PRED(ibbl,iedg)
        {
            pred = AEDG_HEAD(iedg);
            /* this is a quick hack --- NEEDS WORK */
            
            if( AEDG_TYPE(iedg) == ET_HELL || AEDG_TYPE(iedg) == ET_COMPENSATE  )
            {
                ABBL_USEDEF_CONFLUENCE(ibbl) |= REG2BIT(reg);
                count++;
                break;
            }
            
            def = ABBL_USEDEF_DEF(pred);
            if( def == USEDEF_BOT )
                def = ABBL_USEDEF_IN(pred);
            if( def == USEDEF_TOP ) continue;
            CERROR( def == USEDEF_BOT,("weird\n"));
            count++;

            ABBL_USEDEF_CONFLUENCE(ibbl) |= REG2BIT(reg);
            break;
        }
    }
#if 0
    printf("reg %d: added %d/%d confluence points\n",reg,count,pot);
#endif
    return flag;
}


/*=======================================================
=======================================================*/
#undef __PROC__
#define __PROC__ "ComUseDefDag"

extern void ComUseDefDag(FILE *fp)
{
    INDEX  iins;
    INDEX  def;
    INDEX  ibbl;
    NUMBER ccount = 0;
    NUMBER icount = 0;
    
    REG    reg;
    REGLIST reglist;
    
    clock_t time = clock();

    reglist = REGS_ALL_BUT_ZERO;

    UseDefInitialConfluencePoints();


    FOREACH_LIVE_INS(iins)
    {
        if( !(AINS_FLAGS(iins) & FI_EVAL_C) )
        {
            if( AINS_REG_C(iins) == REG_ZERO || AINS_REG_C(iins) == REG_FZERO )
            {
                AINS_C_EVAL(iins) = 0L; /* works for int and flt */
                AINS_FLAGS(iins) |= FI_EVAL_C;

            }
        }

        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            if( AINS_REG_B(iins) == REG_ZERO || AINS_REG_B(iins) == REG_FZERO )
            {
                AINS_B_EVAL(iins) = 0L;
                AINS_FLAGS(iins) |= FI_EVAL_B;
            }

            if( AINS_REG_B(iins) == REG_LIT )
            {
                AINS_B_EVAL(iins) = AINS_DATA(iins);
                AINS_FLAGS(iins) |= FI_EVAL_B;
            }
        }

        if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
        {
            if( AINS_REG_A(iins) == REG_ZERO || AINS_REG_A(iins) == REG_FZERO )
            {
                AINS_A_EVAL(iins) = 0L;
                AINS_FLAGS(iins) |= FI_EVAL_A;
            }
        }
    }
    

    
    
    for(reg=0;reg<64;reg++)
    {
#if 0
        printf("Doing register %d %016lx\n",reg,reglist);
        fflush(stdout);
#endif
        
        if( reg == REG_ZERO || reg == REG_FZERO ) continue;
        
        if( REG2BIT(reg) & reglist )
        {
            do
            {
                UseDefPropagation(reg);
            }
            while( UseDefMoreConfluencePoints(reg) );
            
            UseDefPropagateDefsToInstructions(reg);
        }
        else
        {
            FOREACH_LIVE_BBL(ibbl)
            {
                ABBL_USEDEF_IN(ibbl)  = USEDEF_TOP;
                ABBL_USEDEF_CONFLUENCE(ibbl) |= REG2BIT(reg);
                
            }

            UseDefPropagateDefsToInstructions(reg);
        }
    }


    /* ====================================================================== */
    /* collect stats */
    /* ====================================================================== */
    
    FOREACH_LIVE_BBL(ibbl)
    {
        ABBL_USEDEF_CONFLUENCE(ibbl) = REGS_NONE;
    }
        
    FOREACH_LIVE_INS(iins)
    {
        if( AINS_REG_C(iins) != REG_ZERO && AINS_REG_C(iins) != REG_FZERO )
            icount++;
        
        if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
        {
            def = AINS_A_EVAL(iins);
            if( VALID(def) && def > CONFIG_NINS )
            {
                def -= CONFIG_NINS;
                ibbl = def / 64;
                reg  = def % 64;
                ABBL_USEDEF_CONFLUENCE(ibbl) |= REG2BIT(reg);
            }
            
        }
        if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
        {
            def = AINS_B_EVAL(iins);
            if( VALID(def) && def > CONFIG_NINS )
            {
                def -= CONFIG_NINS;
                ibbl = def / 64;
                reg  = def % 64;
                ABBL_USEDEF_CONFLUENCE(ibbl) |= REG2BIT(reg);
            }
        }
        
    }

    FOREACH_LIVE_BBL(ibbl)
    {
        ccount += BinaryBits(ABBL_USEDEF_CONFLUENCE(ibbl) & REGS_ALL_BUT_ZERO);
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;    

    InfoStatus(fp,"\t[ComUsedefdag] instr %d  confluence %d  tot %d{%d ms}\n",icount,ccount,icount+ccount,time);
}

/*=======================================================
@COM
This routine works with extended basic blocks so it is
important to run the Dead Code Remover first

Requires DeadCode removal
Requires Stack Usage Info
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ComUseDefDagLoadStoreExtBbl"

extern void ComUseDefDagLoadStoreExtBbl(FILE *fp)
{
    INDEX ibbl;
    INDEX jbbl;
    INDEX iins;
    INDEX iins_def;
    REG reg_def;
    clock_t time = clock();
  
    FOREACH_LIVE_BBL(ibbl)
    {
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
            {
                AINS_A_EVAL(iins) = INVALID;
            }


            if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                if( AINS_REG_B(iins) == REG_ZERO )
                {
                    AINS_B_EVAL(iins) = 0;
                    AINS_FLAGS(iins) |= FI_EVAL_B;
                }
                else if( AINS_TYPE(iins) == IT_LDA ||
                         AINS_TYPE(iins) == IT_ILD ||
                         AINS_TYPE(iins) == IT_FLD ||
                         AINS_TYPE(iins) == IT_ULD ||
                         AINS_TYPE(iins) == IT_LLD ||
                         AINS_TYPE(iins) == IT_UST ||
                         AINS_TYPE(iins) == IT_FST ||
                         AINS_TYPE(iins) == IT_IST ||
                         AINS_TYPE(iins) == IT_CST )
                {
                    reg_def = AINS_REG_B(iins);
                    iins_def = AinsFindDefInExtBbl(iins,reg_def);
                    if( !VALID(iins_def) )
                    {
                        for( jbbl = ibbl; AbblNpreds(jbbl) == 1; jbbl = AEDG_HEAD(ABBL_PRED(jbbl)) )
                            ;
                        iins_def = CONFIG_NINS + jbbl * 64 + reg_def;
                    }
                    
                    AINS_B_EVAL(iins) = iins_def;
                }
                else
                {
                    AINS_B_EVAL(iins) = INVALID;
                }
            }
        }
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[ComUsedefdagloadstoreextbbl] {%d ms}\n",time);
}



/*=======================================================
@COM
This routine works with locally onbasic blocks
@END
=======================================================*/
#undef __PROC__
#define __PROC__ "ComUseDefDagLoadStoreExtBbl"

extern void ComUseDefDagBbl(FILE *fp)
{
    INDEX ibbl;
    INDEX iins;
    INDEX def[64];
    REG reg;
    clock_t time = clock();
  
    FOREACH_LIVE_BBL(ibbl)
    {
        iins = CONFIG_NINS + 64 * ibbl;
        for( reg = 0; reg < 64; reg++ )
        {
            if( reg == REG_ZERO || reg == REG_FZERO )
            {
                def[reg] = INVALID;
            }
            else
            {
                def[reg] = iins + reg;
            }
        }
        
        FOREACH_BBL_INS(ibbl,iins)
        {
            if( !(AINS_FLAGS(iins) & FI_EVAL_A) )
            {
                reg = AINS_REG_A(iins);
                if( reg == REG_ZERO || reg == REG_FZERO )
                {
                    AINS_A_EVAL(iins) = 0L;
                    AINS_FLAGS(iins) |= FI_EVAL_A;
                }
                else
                {
                    AINS_A_EVAL(iins) = def[reg];
                }
            }

            if( !(AINS_FLAGS(iins) & FI_EVAL_B) )
            {
                reg = AINS_REG_B(iins);
                if( reg == REG_ZERO || reg == REG_FZERO )
                {
                    AINS_B_EVAL(iins) = 0L;
                    AINS_FLAGS(iins) |= FI_EVAL_B;
                }
                else
                {
                    AINS_B_EVAL(iins) = def[reg];
                }
            }

            reg = AINS_REG_C(iins);
            if( reg != REG_ZERO && reg != REG_FZERO )
            {
                def[reg] = iins;
            }
        }
    }
    
    time = (clock() - time) * 1000 / CLOCKS_PER_SEC;

    InfoStatus(fp,"\t[ComUsedefdagbbl] {%d ms}\n",time);
}



/*=======================================================
======================================================= */
